
import unittest

import pytest

from deolingo._rewriting_translator import DeolingoRewritingTranslator
from deolingo._translator import DeolingoTranslator


class DeolingoTranslatorTestCase(unittest.TestCase):
    
    @staticmethod
    def get_sut(callback):
        return DeolingoTranslator(add_to_program_callback=callback, translate=True,
                           add_theory=False, add_deontic_rules=False)
    
    def parse(self, string) -> str:
        translated = ""

        def callback(x):
            nonlocal translated
            translated += str(x) + "\n"

        translator = self.get_sut(callback)
        translator.transform_source(string)
        return translated

    def test_rule(self):
        actual = self.parse("&obligatory{a} :- &forbidden{b}.")
        expected = "deolingo_obligatory(a) :- deolingo_forbidden(b).\n"
        self.assertEqual(expected, actual)

    def test_rule_with_negation(self):
        actual = self.parse("&omissible{a} :- not &forbidden{b}.")
        expected = "deolingo_omissible(a) :- not deolingo_forbidden(b).\n"
        self.assertEqual(expected, actual)

    def test_deontic_sequence_head(self):
        actual = self.parse("&obligatory{a; -b}.")
        expected = "deolingo_obligatory(a); deolingo_obligatory(-b).\n"
        self.assertEqual(expected, actual)

    def test_deontic_sequence_body(self):
        actual = self.parse(":- &obligatory{a; -b}.")
        expected = "#false :- deolingo_obligatory(a); deolingo_obligatory(-b).\n"
        self.assertEqual(expected, actual)

    def test_deontic_sequence_head_variable(self):
        actual = self.parse("&obligatory{a;X} :- p(X).")
        expected = "deolingo_obligatory(a); deolingo_obligatory(X) :- p(X).\n"
        self.assertEqual(expected, actual)

    def test_deontic_sequence_default_negation_body(self):
        actual = self.parse(":- not &obligatory{a; b}.")
        expected = "#false :- not deolingo_obligatory(a); not deolingo_obligatory(b).\n"
        self.assertEqual(expected, actual)

    def test_deontic_sequence_body_variable(self):
        actual = self.parse(":- &obligatory{X;b}, p(X).")
        expected = "#false :- p(X); deolingo_obligatory(X); deolingo_obligatory(b).\n"
        # TODO: preserve order of atoms!
        self.assertEqual(expected, actual)

    def test_deontic_sequence_with_conditions_head(self):
        actual = self.parse("&obligatory{a: b; -c: d}.")
        expected = "deolingo_obligatory(a): b; deolingo_obligatory(-c): d.\n"
        self.assertEqual(expected, actual)

    def test_deontic_sequence_with_conditions_body(self):
        actual = self.parse(":- &obligatory{a: b; -c: d}.")
        expected = "#false :- deolingo_obligatory(a): b; deolingo_obligatory(-c): d.\n"
        self.assertEqual(expected, actual)

    def test_deontic_conditional(self):
        actual = self.parse("&obligatory{a|b}.")
        expected = "deolingo_obligatory(a) :- b.\ndeolingo_obligatory(a) :- deolingo_non_violated_obligation(b).\n"
        self.assertEqual(expected, actual)

    def test_deontic_conditional_explicit_negation(self):
        actual = self.parse("&obligatory{-a | -b}.")
        expected = "deolingo_obligatory(-a) :- -b.\ndeolingo_obligatory(-a) :- deolingo_non_violated_obligation(-b).\n"
        self.assertEqual(expected, actual)

    def test_deontic_conditional_variable(self):
        actual = self.parse("&obligatory{X | a} :- p(X).")
        expected = "deolingo_obligatory(X) :- a; p(X).\ndeolingo_obligatory(X) :- deolingo_non_violated_obligation(a); p(X).\n"
        self.assertEqual(expected, actual)

    def test_disjunction_head(self):
        actual = self.parse("&obligatory{a || b}.")
        expected = "deolingo_obligatory(a); deolingo_obligatory(b).\n"
        self.assertEqual(expected, actual)

    def test_disjunction_body(self):
        with pytest.raises(Exception):
            self.parse(":- &obligatory{a || b}.")

    def test_conjunction_head(self):
        with pytest.raises(Exception):
            self.parse("&obligatory{a && b}.")

    def test_conjunction_body(self):
        actual = self.parse(":- &obligatory{a && b}.")
        expected = "#false :- deolingo_obligatory(a); deolingo_obligatory(b).\n"
        self.assertEqual(expected, actual)


class DeolingoRewritingTranslatorTestCase(unittest.TestCase):

    @staticmethod
    def get_sut(callback):
        return DeolingoRewritingTranslator(add_to_program_callback=callback, translate=True,
                                           add_theory=False, add_deontic_rules=False)


if __name__ == '__main__':
    unittest.main()




from xclingo import XclingoControl

import re

import clingo.ast as ast

from deolingo._translator import DeolingoTranslator
from deolingo._deontic_atom import _DEOLINGO_ATOM_PREFIX


def rewrite_program(program):
    lines = program.splitlines()
    for i, line in enumerate(lines):
        if line.strip().startswith('%!'):
            repl = _DEOLINGO_ATOM_PREFIX + r"\1(\2)"
            lines[i] = re.sub(r"&(\w+){(.+)}", repl, line)
    return '\n'.join(lines)


class XDeolingoControl(XclingoControl):

    def __init__(self, n_solutions='1', n_explanations='1', auto_trace='none', weak=False):
        super().__init__(n_solutions=n_solutions, n_explanations=n_explanations, auto_trace=auto_trace)
        self.rewritten_program = ""
        self._transformer = DeolingoTranslator(self._add_to_xcontrol, translate=True, add_theory=False, weak=weak)

    def add(self, name, parameters, program):
        translated_part = self._transformer.transform_source(program)
        self._add_translated(translated_part)

    def add_inputs(self, inputs):
        self._transformer.transform_sources(inputs)
        xprogram = self._transformer.translated_program
        self._add_translated(xprogram)

    def _add_translated(self, xprogram):
        rewritten_program = rewrite_program(xprogram)
        self.rewritten_program = rewritten_program
        super().add("base", [], rewritten_program)

    def _add_to_xcontrol(self, statement: ast.AST) -> None:
        pass




import sys
from typing import Callable

import clingo

import deolingo._version as deolingo_version
from deolingo._answer_set_rewriter import DeonticAnswerSetRewriter


class DeolingoApplication(clingo.Application):
    """
    Application object as accepted by clingo.clingo_main().
    Rewrites the incoming deontic logic programs into deontic ASP programs and solves them.
    """

    # <editor-fold desc="Initialization">

    def __init__(self):
        """
        Initializes the application setting the program name. See clingo.clingo_main().
        """
        self.program_name = "deolingo"
        self.version = deolingo_version.__version__
        self._translate_flag = clingo.Flag(False)
        self._ungrouped_flag = clingo.Flag(False)
        self._explain_flag = clingo.Flag(False)
        self._benchmark_flag = clingo.Flag(False)
        self._generate_flag = clingo.Flag(False)
        self._optimize_flag = clingo.Flag(False)
        self._weak_flag = clingo.Flag(False)
        self._generator = None
        self._llm = None
        self._n_explanations = 0
        self._temporal_flag = clingo.Flag(False)
        self._answer_set_rewriter = DeonticAnswerSetRewriter()
        self._xcontrol = None
        self._set_output_format_if_translating()

    # </editor-fold>

    # <editor-fold desc="clingo.Application override">

    def main(self, program, files):
        """This function implements the Application.main() function as required by clingo.clingo_main()."""
        if self._temporal_flag.flag:
            return self._solve_temporal_deontic_program(program, files)
        if self._benchmark_flag.flag:
            return self._run_benchmark()
        if self._generate_flag.flag:
            inputs = self._read_source_inputs_from_files(files)
            return self._generate_deontic_program(inputs)
        if self._explain_flag.flag:
            inputs = self._read_source_inputs_from_files(files)
            return self._explain_deontic_program(program, inputs)
        return self._solve_deontic_program(program, files)

    def register_options(self, options: clingo.ApplicationOptions):
        """Registers the options for the application."""
        options.add_flag("deontic",
                         "translate",
                         "Translate a deontic logic program into an ASP program",
                         self._translate_flag)
        options.add_flag("deontic",
                         "ungrouped",
                         "Do not group the answer sets deontic worlds",
                         self._ungrouped_flag)
        options.add_flag("deontic",
                         "explain",
                         "Use Xclingo to generate explanations",
                         self._explain_flag)
        options.add_flag("deontic",
                         "benchmark",
                         "Run benchmark with all examples and print the results",
                         self._benchmark_flag)
        options.add_flag("deontic",
                         "generate",
                         "Generates a deontic logic program from a given natural language input",
                         self._generate_flag)
        options.add_flag("deontic",
                         "optimize",
                         "Optimize the translation of the deontic logic program",
                         self._optimize_flag)
        options.add_flag("deontic",
                         "weak",
                         "Makes the deontic weak axiom a weak constraint to find inconsistencies",
                         self._weak_flag)

        def parser(value: str):
            from deolingo._generator import Generator
            try:
                self._generator = Generator(value)
                return True
            except ValueError:
                return False

        def llm_parser(value: str):
            self._llm = value
            return True

        def xparser(value: str):
            self._n_explanations = int(value)
            return True
        options.add("deontic", "generator", "Generative AI API to use in --generate mode",
                    parser, False, "gemini|gpt4all|openai|huggingfacehub")
        options.add("deontic", "llm", "Generative AI LLM model to be used in --generate mode",
                    llm_parser, False, "gemini-1.5-flash-latest")
        options.add("deontic", "explanations", "Number of explanations to generate",
                    xparser, False, "0..N")
        options.add_flag("deontic",
                         "temporal",
                         "Runs a temporal deontic logic program in Telingo",
                         self._temporal_flag)

    def print_model(self, model: clingo.Model, printer: Callable[[], None] = None):
        """Prints the atoms of the given model.
        This function is called for each model of the problem."""
        atoms = model.symbols(shown=True)
        grouped = not self._ungrouped_flag.flag
        self._answer_set_rewriter._grouped = grouped
        rewritten_facts, obligations, prohibitions = self._answer_set_rewriter.rewrite_atoms(atoms)
        if grouped:
            print(f'FACTS: {", ".join(rewritten_facts)}')
            print(f'OBLIGATIONS: {", ".join(obligations)}')
            print(f'PROHIBITIONS: {", ".join(prohibitions)}')
            print()
        else:
            print(", ".join(rewritten_facts))

    # </editor-fold>

    # <editor-fold desc="Private methods">

    @staticmethod
    def _set_output_format_if_translating():
        if "--translate" not in sys.argv and "--temporal" not in sys.argv:
            return
        for arg in sys.argv:
            if arg.startswith("--outf="):
                sys.argv.remove(arg)
        sys.argv.append("--outf=3")

    @staticmethod
    def _read_source_inputs_from_files(files):
        files = [open(file) for file in files]
        if len(files) == 0:
            files.append(sys.stdin)
        inputs = [file.read() for file in files]
        return inputs

    @staticmethod
    def _run_benchmark():
        from deolingo.domain.run_benchmark_command import RunBenchmarkCommand
        command = RunBenchmarkCommand()
        command.execute()

    def _generate_deontic_program(self, inputs):
        from deolingo.domain.generate_deontic_program_command import GenerateDeonticProgramCommand
        command = GenerateDeonticProgramCommand(inputs[0], generator_type=self._generator, llm=self._llm)
        command.execute()

    def _solve_deontic_program(self, program, files):
        from deolingo.domain.solve_deontic_program_command import SolveDeonticProgramCommand
        command = SolveDeonticProgramCommand(program, files, self._translate_flag.flag, self._optimize_flag.flag,
                                             self._weak_flag.flag)
        command.execute()

    def _explain_deontic_program(self, program, inputs, n_solutions='1'):
        from deolingo.domain.explain_deontic_program_command import ExplainDeonticProgramCommand
        n_solutions = int(program.configuration.solve.models)
        n_solutions_str = '1' if n_solutions < 0 else str(n_solutions)
        command = ExplainDeonticProgramCommand(inputs, n_solutions_str, self._n_explanations, self._translate_flag.flag,
                                               self._weak_flag.flag)
        command.execute()

    def _solve_temporal_deontic_program(self, program, files):
        from deolingo.domain.solve_temporal_deontic_program_command import SolveTemporalDeonticProgramCommand
        command = SolveTemporalDeonticProgramCommand(program, files, self._weak_flag.flag)
        command.execute()

    # </editor-fold>




from deolingo.control import DeolingoControl
from deolingo.rewriting_control import DeolingoRewritingControl


class DeolingoSolver:

    def __init__(self, all_models=False, grouped=False, control=None, optimize=False):
        self._control = control if control is not None \
            else DeolingoRewritingControl(grouped=grouped) if optimize \
            else DeolingoControl(grouped=grouped)
        self._all_models = all_models
        self._models = []
        if all_models:
            self._control.configuration.solve.models = 0

    def solve(self, program):
        self._control.add(program)
        self._control.ground()
        self._control.solve(on_model=self._on_model)
        return self._models

    def _on_model(self, model):
        self._models.append(model)
        return model



import clingo

from deolingo._deontic_atom import DeonticAtoms, DeonticAtom


class DeonticAnswerSetRewriter:

    # <editor-fold desc="Initialization">

    def __init__(self, grouped=False):
        self._grouped = grouped

    # </editor-fold>

    # <editor-fold desc="Public methods">

    def rewrite_atoms(self, atoms):
        """Rewrites the atom by removing the prefix 'deolingo_' if present."""
        rewritten_atoms = []
        for atom in atoms:
            str_atom = str(atom)
            is_negated = str_atom.startswith("-")
            deontic_atom = DeonticAtoms.with_prefixed_name(atom.name)
            if deontic_atom is None:
                if self._grouped:
                    rewritten_atoms.append((str_atom, None))
                else:
                    rewritten_atoms.append(str_atom)
                continue
            deontic_atoms = [DeonticAtoms.OBLIGATORY, DeonticAtoms.FORBIDDEN,
                             DeonticAtoms.PERMITTED, DeonticAtoms.OMISSIBLE]
            if is_negated or deontic_atom not in deontic_atoms:
                continue
            str_atom = str_atom[1:] if is_negated else str_atom
            unprefixed = DeonticAtom.unprefix(str_atom)
            # remove the deontic atom name and the first parenthesis from str_atom
            str_atom_int = unprefixed.replace(deontic_atom.value.name, "", 1)[1:]
            is_negated_inside = str_atom_int.startswith("-")
            if is_negated_inside:
                continue
            rewritten = "&" + unprefixed.replace("(", "{", 1)
            rewritten = rewritten[::-1].replace(")", "}", 1)[::-1]
            rewritten = rewritten if not is_negated else "-" + rewritten
            if self._grouped:
                rewritten_atoms.append((rewritten, deontic_atom))
            else:
                rewritten_atoms.append(rewritten)
        if self._grouped:
            return self._group_atoms_by_worlds(rewritten_atoms)
        else:
            return rewritten_atoms, None, None

    def rewrite_model(self, model: clingo.Model):
        """Rewrites the atoms of the given model by removing the prefix '_deolingo_' if present."""
        atoms = model.symbols(shown=True)
        rewritten_atoms, obligations, prohibitions = self.rewrite_atoms(atoms)
        if not self._grouped:
            return rewritten_atoms
        return {'facts': list(rewritten_atoms),
                'obligations': list(obligations),
                'prohibitions': list(prohibitions)}


    # </editor-fold>

    # <editor-fold desc="Private methods">

    @staticmethod
    def _group_atoms_by_worlds(rewritten_atoms):
        facts_world = set()
        obligations_world = set()
        prohibitions_world = set()
        for atom_str, deontic_atom in rewritten_atoms:
            if deontic_atom in [DeonticAtoms.OBLIGATORY, DeonticAtoms.OMISSIBLE]:
                obligations_world.add(atom_str)
            elif deontic_atom in [DeonticAtoms.FORBIDDEN, DeonticAtoms.PERMITTED]:
                prohibitions_world.add(atom_str)
            else:
                facts_world.add(atom_str)
        return facts_world, obligations_world, prohibitions_world

    # </editor-fold>



import unittest

from deolingo.examples import ExampleReader
from deolingo.solver import DeolingoSolver


class DeolingoDELXExamplesTestCase(unittest.TestCase):

    @staticmethod
    def get_sut(all_models=False):
        return DeolingoSolver(all_models)

    def test_delx_example_1_1(self):
        # Arrange
        example = ExampleReader().read_example("delx/example1.1.lp")
        expected_answer_set = {'&forbidden{f}'}
        # Act
        actual_answer_sets = self.get_sut().solve(example.contents)
        # Assert
        assert len(actual_answer_sets) == 1
        actual_answer_set = set(actual_answer_sets[0])
        assert actual_answer_set == expected_answer_set

    def test_delx_example_1_2(self):
        # Arrange
        example = ExampleReader().read_example("delx/example1.2.lp")
        expected_answer_set = {'&forbidden{f}'}
        # Act
        actual_answer_sets = self.get_sut().solve(example.contents)
        # Assert
        assert len(actual_answer_sets) == 1
        actual_answer_set = set(actual_answer_sets[0])
        assert actual_answer_set == expected_answer_set

    def test_delx_example_2(self):
        # Arrange
        example = ExampleReader().read_example("delx/example2.lp")
        expected_answer_set = {'&forbidden{f}'}
        # Act
        actual_answer_sets = self.get_sut().solve(example.contents)
        # Assert
        assert len(actual_answer_sets) == 1
        actual_answer_set = set(actual_answer_sets[0])
        assert actual_answer_set == expected_answer_set

    def test_delx_example_3(self):
        # Arrange
        example = ExampleReader().read_example("delx/example3.lp")
        expected_answer_set = {'f', '&obligatory{w}', '&obligatory{f}', '&obligatory{pay}', '&forbidden{f}'}
        # Act
        actual_answer_sets = self.get_sut().solve(example.contents)
        # Assert
        assert len(actual_answer_sets) == 1
        actual_answer_set = set(actual_answer_sets[0])
        assert actual_answer_set == expected_answer_set

    def test_delx_example_4(self):
        # Arrange
        example = ExampleReader().read_example("delx/example4.lp")
        expected_answer_set = {'&permitted{f}', 'f', 's'}
        # Act
        actual_answer_sets = self.get_sut().solve(example.contents)
        # Assert
        assert len(actual_answer_sets) == 1
        actual_answer_set = set(actual_answer_sets[0])
        assert actual_answer_set == expected_answer_set

    def test_delx_example_5_1(self):
        # Arrange
        example = ExampleReader().read_example("delx/example5.1.lp")
        expected_answer_set = {'&forbidden{f}', '&forbidden{w}', 'f', '&obligatory{w}', '&obligatory{f}', 'w'}
        # Act
        actual_answer_sets = self.get_sut().solve(example.contents)
        # Assert
        assert len(actual_answer_sets) == 1
        actual_answer_set = set(actual_answer_sets[0])
        assert actual_answer_set == expected_answer_set

    def test_delx_example_5_2(self):
        # Arrange
        example = ExampleReader().read_example("delx/example5.2.lp")
        expected_answer_set = {'f', '&obligatory{w}', '&obligatory{f}', '&forbidden{f}'}
        # Act
        actual_answer_sets = self.get_sut().solve(example.contents)
        # Assert
        assert len(actual_answer_sets) == 1
        actual_answer_set = set(actual_answer_sets[0])
        assert actual_answer_set == expected_answer_set

    def test_delx_example_6_1(self):
        # Arrange
        example = ExampleReader().read_example("delx/example6.1.lp")
        expected_answer_sets = [
            # Model 1
            {'-f', '&forbidden{m}', '&forbidden{f}'},
            # Model 2
            {'f', '&obligatory{w}', '&obligatory{m}', '&obligatory{f}', '&forbidden{f}'}
        ]
        # Act
        actual_answer_sets = self.get_sut(all_models=True).solve(example.contents)
        # Assert
        assert len(actual_answer_sets) == 2
        actual_answer_set_0 = set(actual_answer_sets[0])
        actual_answer_set_1 = set(actual_answer_sets[1])
        assert actual_answer_set_0 == expected_answer_sets[0]
        assert actual_answer_set_1 == expected_answer_sets[1]

    def test_delx_example_6_2(self):
        # Arrange
        example = ExampleReader().read_example("delx/example6.2.lp")
        expected_answer_sets = [
            # Model 1
            {'-f', '&forbidden{m}', '&forbidden{f}'},
            # Model 2
            {'f', '&obligatory{w}', '&obligatory{m}', '&obligatory{f}', '&forbidden{f}'}
        ]
        # Act
        actual_answer_sets = self.get_sut(all_models=True).solve(example.contents)
        # Assert
        assert len(actual_answer_sets) == 2
        actual_answer_set_0 = set(actual_answer_sets[0])
        actual_answer_set_1 = set(actual_answer_sets[1])
        assert actual_answer_set_0 == expected_answer_sets[0]
        assert actual_answer_set_1 == expected_answer_sets[1]


class DeolingoDELXExamplesOptimizedTestCase(DeolingoDELXExamplesTestCase):

    @staticmethod
    def get_sut(all_models=False):
        return DeolingoSolver(all_models=all_models, optimize=True)




import unittest

from deolingo.examples import ExampleReader
from deolingo.solver import DeolingoSolver


class DeolingoPreliminaryExamplesTestCase(unittest.TestCase):

    @staticmethod
    def get_sut():
        return DeolingoSolver()

    def test_preliminary_example_1(self):
        # Arrange
        example = ExampleReader().read_example("preliminary/example1.lp")
        expected_answer_set = {"park"}
        # Act
        actual_answer_sets = self.get_sut().solve(example.contents)
        # Assert
        assert len(actual_answer_sets) == 1
        actual_answer_set = set(actual_answer_sets[0])
        assert actual_answer_set == expected_answer_set
    
    def test_preliminary_example_2(self):
        # Arrange
        example = ExampleReader().read_example("preliminary/example2.lp")
        expected_answer_set = {'&obligatory{work}', '-work', '-weekend'}
        # Act
        actual_answer_sets = self.get_sut().solve(example.contents)
        # Assert
        assert len(actual_answer_sets) == 1
        actual_answer_set = set(actual_answer_sets[0])
        assert actual_answer_set == expected_answer_set
    
    def test_preliminary_example_3(self):
        # Arrange
        example = ExampleReader().read_example("preliminary/example3.lp")
        # Act
        actual_answer_sets = self.get_sut().solve(example.contents)
        # Assert
        assert actual_answer_sets == []

    def test_preliminary_example_4(self):
        # Arrange
        example = ExampleReader().read_example("preliminary/example4.lp")
        expected_answer_set = {'walk', '&obligatory{walk_right}', '&obligatory{walk}', 'walk_right', '&forbidden{walk}'}
        # Act
        actual_answer_sets = self.get_sut().solve(example.contents)
        print(actual_answer_sets)
        # Assert
        assert len(actual_answer_sets) == 1
        actual_answer_set = set(actual_answer_sets[0])
        assert actual_answer_set == expected_answer_set
    
    def test_preliminary_example_5(self):
        # Arrange
        example = ExampleReader().read_example("preliminary/example5.lp")
        expected_answer_set = {'&permitted{fence}', 'fence', 'sea'}
        # Act
        actual_answer_sets = self.get_sut().solve(example.contents)
        # Assert
        assert len(actual_answer_sets) == 1
        actual_answer_set = set(actual_answer_sets[0])
        assert actual_answer_set == expected_answer_set


class DeolingoPreliminaryExamplesOptimizedTestCase(DeolingoPreliminaryExamplesTestCase):

    @staticmethod
    def get_sut():
        return DeolingoSolver(optimize=True)




from clingo.ast import parse_string, parse_files

from deolingo._ast_transformer import DeonticASTTransformer, MultipleRuleException

from deolingo._deontic_atom import *
from deolingo._deolingo_theory import _DEOLINGO_THEORY
from deolingo._deontic_rules import DeonticRules


class DeolingoTranslator:

    # <editor-fold desc="Initialization">

    def __init__(self, add_to_program_callback, transformer=None, translate=False, add_theory=True,
                 add_deontic_rules=True, temporal=False, weak=False):
        self.deontic_transformer = transformer if transformer is not None else DeonticASTTransformer(translate)
        self._add_to_program_callback = add_to_program_callback
        self.translate = translate
        self.translated_program = ""
        self._translated_part = ""
        self._deolingo_theory_added = False
        self._deontic_rules_added = False
        self._add_theory = add_theory
        self._add_deontic_rules = add_deontic_rules
        self.temporal = temporal
        self.weak = weak

    # </editor-fold>

    # <editor-fold desc="Public methods">

    def transform_source(self, source):
        """Transforms the source input and adds it to the program."""
        return self.transform_sources([source])

    def transform_sources(self, inputs, files=None):
        """Transforms the source inputs and adds them to the program."""
        self._translated_part = ""
        self._add_deolingo_theory()
        self._transform_and_add_source_inputs(inputs, files)
        self._add_program_statement()
        self._add_common_deontic_rules()
        self._add_rules_for_each_deontic_atom()
        part = self._translated_part
        self._translated_part = ""
        return part

    # </editor-fold>

    # <editor-fold desc="Private methods">

    def _add_to_translation(self, statement):
        """Add the statement to the translated program if necessary."""
        if self.translate:
            statement_str = str(statement)
            self.translated_program += statement_str
            self._translated_part += statement_str

    def _add_string_to_program(self, statement, add_to_translation=True):
        """Add the statement to the program and to the translated program if necessary."""
        parse_string(statement, self._add_to_program_callback)
        if add_to_translation:
            self._add_to_translation(statement)

    def _add_deolingo_theory(self):
        """Add the Deolingo theory to the program if it has not been added yet."""
        if self._add_theory and not self._deolingo_theory_added:
            self._add_string_to_program(_DEOLINGO_THEORY, add_to_translation=self.translate)
            self._deolingo_theory_added = True

    def _transform_and_add_to_program(self, statement):
        """Transform the source statement and add it to the program."""
        try:
            self.deontic_transformer.translated_program = ""
            transformed_statement = self.deontic_transformer(statement)
            if str(transformed_statement) == "#program base.":
                return
            if self.deontic_transformer.translated_program == "":
                self._add_to_translation("\n" + str(transformed_statement))
            else:
                self._add_to_translation(self.deontic_transformer.translated_program)
            self._add_to_program_callback(transformed_statement)
        except MultipleRuleException as multiple_rule:
            self._add_multiple_rules_to_program(multiple_rule)

    def _add_multiple_rules_to_program(self, multiple_rule):
        if multiple_rule.line is not None:
            self.translated_program += f"\n% Source line: {multiple_rule.line.begin.line}\n"
        for rule in multiple_rule.rules:
            self._add_to_translation(str(rule) + "\n")
            self._add_to_program_callback(rule)

    def _transform_and_add_source_inputs(self, inputs, files=None):
        """Parses the source inputs, transforms them and adds them to the program."""
        if files is not None:
            parse_files(files, self._transform_and_add_to_program)
            return
        for source_input in inputs:
            parse_string(source_input, self._transform_and_add_to_program)

    def _add_rules_for_each_deontic_atom(self):
        """Add the reified truth rules for each deontic atom."""
        if not self._add_deontic_rules:
            return
        for deontic_atom in self.deontic_transformer.deontic_atoms:
            holds_positive = holds(deontic_atom)
            holds_negative = holds(f"-{deontic_atom}")
            rules = [
                f"\n% Deontic atom rules for '{deontic_atom}'",
                f"{holds_positive} :- {deontic_atom}.",
                f"{holds_negative} :- -{deontic_atom}.",
                f"{deontic_atom} :- {holds_positive}.",
                f"-{deontic_atom} :- {holds_negative}."
            ]
            rules_as_string = "\n".join(rules) + '\n'
            self._add_string_to_program(rules_as_string)

    def _add_program_statement(self):
        if self._add_deontic_rules and self.translate:
            self._transform_and_add_source_inputs(["#program base."])
        if self._add_deontic_rules and self.temporal:
            self._transform_and_add_source_inputs(["#program always."])

    def _add_common_deontic_rules(self):
        """Add the common deontic rules to the program if not added already."""
        if not self._add_deontic_rules or self._deontic_rules_added:
            return
        deontic_rules_as_string = DeonticRules.all_rules_as_string(self.weak)
        self._add_string_to_program(deontic_rules_as_string)
        self._deontic_rules_added = True

    # </editor-fold>




import unittest

from deolingo._ast_rewriting_transformer import DeonticASTRewritingTransformer
from clingo.ast import parse_string


def parse(string) -> str:
    transformer = DeonticASTRewritingTransformer(translate=True)
    translated = ""

    def callback(x):
        nonlocal translated
        translated = str(transformer(x))
    parse_string(string, callback)
    return translated


class DeolingoRewritingTranslatorTestCase(unittest.TestCase):

    def test_permitted_implicitly(self):
        actual = parse(":- &permitted_implicitly{a}.")
        expected = "#false :- not deolingo_forbidden(a)."
        self.assertEqual(expected, actual)

    def test_permitted_implicitlyg_negated(self):
        actual = parse(":- not &permitted_implicitly{a}.")
        expected = "#false :- not not deolingo_forbidden(a)."
        self.assertEqual(expected, actual)

    def test_permitted_implicitly_strong_negated(self):
        actual = parse(":- &permitted_implicitly{-a}.")
        expected = "#false :- not deolingo_forbidden(-a)."
        self.assertEqual(expected, actual)

    def test_omissible_implicitly(self):
        actual = parse(":- &omissible_implicitly{a}.")
        expected = "#false :- not deolingo_obligatory(a)."
        self.assertEqual(expected, actual)

    def test_omissible_implicitlyg_negated(self):
        actual = parse(":- not &omissible_implicitly{a}.")
        expected = "#false :- not not deolingo_obligatory(a)."
        self.assertEqual(expected, actual)

    def test_omissible_implicitly_strong_negated(self):
        actual = parse(":- &omissible_implicitly{-a}.")
        expected = "#false :- not deolingo_obligatory(-a)."
        self.assertEqual(expected, actual)

    def test_violated_obligation(self):
        actual = parse(":- &violated_obligation{a}.")
        expected = "#false :- deolingo_obligatory(a); deolingo_holds(-a)."
        self.assertEqual(expected, actual)

    def test_violated_obligation_negated(self):
        actual = parse(":- not &violated_obligation{a}.")
        expected = "#false :- not deolingo_violated_obligation(a)."
        self.assertEqual(expected, actual)

    def test_violated_obligation_strong_negated(self):
        actual = parse(":- &violated_obligation{-a}.")
        expected = "#false :- deolingo_obligatory(-a); deolingo_holds(a)."
        self.assertEqual(expected, actual)

    def test_non_violated_obligation(self):
        actual = parse(":- &non_violated_obligation{a}.")
        expected = "#false :- deolingo_obligatory(a); not deolingo_holds(-a)."
        self.assertEqual(expected, actual)

    def test_non_violated_obligation_strong_negated(self):
        actual = parse(":- &non_violated_obligation{-a}.")
        expected = "#false :- deolingo_obligatory(-a); not deolingo_holds(a)."
        self.assertEqual(expected, actual)

    def test_fulfilled_obligation(self):
        actual = parse(":- &fulfilled_obligation{a}.")
        expected = "#false :- deolingo_obligatory(a); deolingo_holds(a)."
        self.assertEqual(expected, actual)

    def test_fulfilled_obligation_strong_negated(self):
        actual = parse(":- &fulfilled_obligation{-a}.")
        expected = "#false :- deolingo_obligatory(-a); deolingo_holds(-a)."
        self.assertEqual(expected, actual)

    def test_non_fulfilled_obligation(self):
        actual = parse(":- &non_fulfilled_obligation{a}.")
        expected = "#false :- deolingo_obligatory(a); not deolingo_holds(a)."
        self.assertEqual(expected, actual)

    def test_non_fulfilled_obligation_strong_negated(self):
        actual = parse(":- &non_fulfilled_obligation{-a}.")
        expected = "#false :- deolingo_obligatory(-a); not deolingo_holds(-a)."
        self.assertEqual(expected, actual)

    def test_undetermined_obligation(self):
        actual = parse(":- &undetermined_obligation{a}.")
        expected = "#false :- deolingo_obligatory(a); not deolingo_holds(a); not deolingo_holds(-a)."
        self.assertEqual(expected, actual)

    def test_undetermined_obligation_strong_negated(self):
        actual = parse(":- &undetermined_obligation{-a}.")
        expected = "#false :- deolingo_obligatory(-a); not deolingo_holds(-a); not deolingo_holds(a)."
        self.assertEqual(expected, actual)

    def test_violated_prohibition(self):
        actual = parse(":- &violated_prohibition{a}.")
        expected = "#false :- deolingo_forbidden(a); deolingo_holds(a)."
        self.assertEqual(expected, actual)

    def test_violated_prohibition_strong_negated(self):
        actual = parse(":- &violated_prohibition{-a}.")
        expected = "#false :- deolingo_forbidden(-a); deolingo_holds(-a)."
        self.assertEqual(expected, actual)

    def test_non_violated_prohibition(self):
        actual = parse(":- &non_violated_prohibition{a}.")
        expected = "#false :- deolingo_forbidden(a); not deolingo_holds(a)."
        self.assertEqual(expected, actual)

    def test_non_violated_prohibition_strong_negated(self):
        actual = parse(":- &non_violated_prohibition{-a}.")
        expected = "#false :- deolingo_forbidden(-a); not deolingo_holds(-a)."
        self.assertEqual(expected, actual)

    def test_fulfilled_prohibition(self):
        actual = parse(":- &fulfilled_prohibition{a}.")
        expected = "#false :- deolingo_forbidden(a); deolingo_holds(-a)."
        self.assertEqual(expected, actual)

    def test_fulfilled_prohibition_strong_negated(self):
        actual = parse(":- &fulfilled_prohibition{-a}.")
        expected = "#false :- deolingo_forbidden(-a); deolingo_holds(a)."
        self.assertEqual(expected, actual)

    def test_non_fulfilled_prohibition(self):
        actual = parse(":- &non_fulfilled_prohibition{a}.")
        expected = "#false :- deolingo_forbidden(a); not deolingo_holds(-a)."
        self.assertEqual(expected, actual)

    def test_non_fulfilled_prohibition_strong_negated(self):
        actual = parse(":- &non_fulfilled_prohibition{-a}.")
        expected = "#false :- deolingo_forbidden(-a); not deolingo_holds(a)."
        self.assertEqual(expected, actual)


if __name__ == '__main__':
    unittest.main()




import time

import prettytable.prettytable

from deolingo.examples import ExampleReader

from deolingo.control import DeolingoControl
from deolingo.rewriting_control import DeolingoRewritingControl


class BenchmarkRunner:
    """
    Class that runs benchmarks for the Deolingo system using all the examples in the examples folder.
    It measures the time it takes to translate the deontic logic programs into ASP programs and to solve them.
    It also measures the Clingo grounding size in terms of the number of ground atoms and the number of ground rules.
    """

    def __init__(self):
        self.solvers = [DeolingoControl, DeolingoRewritingControl]
        self._example_reader = ExampleReader()
        table = prettytable.PrettyTable()
        table.field_names = ["Example", "Solver", "Time (s)", "CPU Time (s)", "Rules", "Atoms"]
        table.align["Example"] = "l"
        table.align["Solver"] = "l"
        table.align["Time (s)"] = "r"
        table.align["CPU Time (s)"] = "r"
        table.align["Rules"] = "r"
        table.align["Atoms"] = "r"
        self._table = table

    def run_benchmark(self):
        """Run the benchmark with all examples and print the results."""
        examples = self._example_reader.read_examples()
        for example in examples:
            if example.name.startswith("telingo/"):
                continue
            is_first_solver = True
            for solver in self.solvers:
                print(f"Running benchmark for {example.name} with {solver.__name__}...")
                self.run_benchmark_for_example(example, solver, is_first_solver)
                is_first_solver = False
            if len(self.solvers) > 1:
                self._table.add_row(["" for _ in range(len(self._table.field_names))])

    def print_results(self):
        """Print the results of the benchmark."""
        print(self._table)

    def run_benchmark_for_example(self, example, solver, is_first_solver):
        """Run the benchmark for the given example and print the results."""
        program = solver()
        if hasattr(program, "configuration"):
            program.configuration.solve.statistics = 2
            program.configuration.solve.models = 0
        program.add("base", [], example.contents)
        program.ground([("base", [])])
        start = time.time()
        program.solve()
        end = time.time()
        print(f"Time: {end - start:.2f} seconds")
        stats = program.statistics
        self._table.add_row([
            example.name if is_first_solver else "",
            solver.__name__,
            f"{end - start:.2f}",
            f"{stats['summary']['times']['cpu']:.2f}",
            int(stats['problem']['lp']['rules']),
            int(stats['problem']['lp']['atoms'])
        ])




from deolingo._telingo_app import DeolingoTelingoApp
from deolingo.domain.use_case_command import UseCaseCommand


class SolveTemporalDeonticProgramCommand(UseCaseCommand):

    def __init__(self, program, files, weak=False):
        super().__init__()
        self.program = program
        self.files = files
        self.weak = weak
        self._app = DeolingoTelingoApp()

    def execute(self):
        self._app.run(self.program, self.files, weak=self.weak)



from copy import deepcopy
from enum import Enum
from dataclasses import dataclass


_DEOLINGO_ATOM_PREFIX = "deolingo_"


@dataclass
class DeonticAtom:

    short_name: str
    long_name: str
    is_tel_next: bool = False
    is_tel_prev: bool = False
    is_tel_init: bool = False

    @property
    def name(self):
        return self.long_name

    def prefixed(self):
        pre_name = self.prefix(self.name)
        if self.is_tel_prev:
            return "'" + pre_name
        elif self.is_tel_next:
            return pre_name + "'"
        elif self.is_tel_init:
            return "_" + pre_name
        return pre_name

    def prefixed_short_name(self):
        return self.prefix(self.short_name)

    def prefixed_long_name(self):
        return self.prefix(self.long_name)

    def wrap(self, atom):
        return f"{self.prefixed()}({atom})"

    @staticmethod
    def prefix(atom):
        """Returns a string representing the prefix for the atoms generated by deolingo."""
        return _DEOLINGO_ATOM_PREFIX + atom

    @staticmethod
    def unprefix(atom):
        """Returns the atom without the prefix 'deolingo_'."""
        return atom[len(_DEOLINGO_ATOM_PREFIX):]


class DeonticAtoms(Enum):
    """
    Enumeration of the deontic atoms.
    """
    OBLIGATORY = DeonticAtom("ob", "obligatory")
    FORBIDDEN = DeonticAtom("fb", "forbidden")
    OMISSIBLE = DeonticAtom("om", "omissible")
    PERMITTED = DeonticAtom("pm", "permitted")
    OPTIONAL = DeonticAtom("op", "optional")
    PERMITTED_BY_DEFAULT = DeonticAtom("pm_d", "permitted_by_default")
    OMISSIBLE_BY_DEFAULT = DeonticAtom("om_d", "omissible_by_default")
    HOLDS = DeonticAtom("h", "holds")
    DEONTIC = DeonticAtom("deo", "deontic")
    PERMITTED_IMPLICITLY = DeonticAtom("pm_i", "permitted_implicitly")
    OMISSIBLE_IMPLICITLY = DeonticAtom("om_i", "omissible_implicitly")
    VIOLATED = DeonticAtom("viol", "violated")
    FULFILLED = DeonticAtom("ful", "fulfilled")
    VIOLATED_OBLIGATION = DeonticAtom("ob_v", "violated_obligation")
    FULFILLED_OBLIGATION = DeonticAtom("ob_f", "fulfilled_obligation")
    NON_VIOLATED_OBLIGATION = DeonticAtom("ob_nv", "non_violated_obligation")
    NON_FULFILLED_OBLIGATION = DeonticAtom("ob_nf", "non_fulfilled_obligation")
    UNDETERMINED_OBLIGATION = DeonticAtom("ob_u", "undetermined_obligation")
    DEFAULT_OBLIGATION = DeonticAtom("ob_d", "default_obligation")
    VIOLATED_PROHIBITION = DeonticAtom("fb_v", "violated_prohibition")
    FULFILLED_PROHIBITION = DeonticAtom("fb_f", "fulfilled_prohibition")
    NON_VIOLATED_PROHIBITION = DeonticAtom("fb_nv", "non_violated_prohibition")
    NON_FULFILLED_PROHIBITION = DeonticAtom("fb_nf", "non_fulfilled_prohibition")
    UNDETERMINED_PROHIBITION = DeonticAtom("fb_u", "undetermined_prohibition")
    DEFAULT_PROHIBITION = DeonticAtom("fb_d", "default_prohibition")

    @classmethod
    def get_all(cls):
        """Returns all the deontic atoms from the enum."""
        return [atom for atom in DeonticAtoms]

    @classmethod
    def get_all_names(cls):
        """Returns all the names of the deontic atoms from the enum."""
        return [atom.value.name for atom in cls.get_all()]

    @classmethod
    def with_name(cls, name):
        """Returns the deontic atom with the given name."""
        is_tel_prev = name.startswith("'")
        is_tel_next = name.endswith("'")
        is_tel_init = name.startswith("_")
        if is_tel_prev or is_tel_next:
            name = name.replace("'", "")
        if is_tel_init:
            name = name.lstrip("_")
        for atom in cls.get_all():
            if atom.value.name == name:
                new_atom = deepcopy(atom.value)
                new_atom.is_tel_prev = is_tel_prev
                new_atom.is_tel_next = is_tel_next
                new_atom.is_tel_init = is_tel_init
                return new_atom
        return None

    @classmethod
    def with_prefixed_name(cls, name):
        """Returns the deontic atom with the given prefixed name."""
        for atom in cls.get_all():
            if atom.value.prefixed() == name:
                return atom
        return None


def deontic(deontic_atom):
    return DeonticAtoms.DEONTIC.value.wrap(deontic_atom)


def violated(deontic_atom):
    return DeonticAtoms.VIOLATED.value.wrap(deontic_atom)


def fulfilled(deontic_atom):
    return DeonticAtoms.FULFILLED.value.wrap(deontic_atom)


def obligatory(deontic_atom):
    return DeonticAtoms.OBLIGATORY.value.wrap(deontic_atom)


def forbidden(deontic_atom):
    return DeonticAtoms.FORBIDDEN.value.wrap(deontic_atom)


def optional(deontic_atom):
    return DeonticAtoms.OPTIONAL.value.wrap(deontic_atom)


def holds(deontic_atom):
    return DeonticAtoms.HOLDS.value.wrap(deontic_atom)


def permitted_implicitly(deontic_atom):
    return DeonticAtoms.PERMITTED_IMPLICITLY.value.wrap(deontic_atom)


def omissible_implicitly(deontic_atom):
    return DeonticAtoms.OMISSIBLE_IMPLICITLY.value.wrap(deontic_atom)


def omissible(deontic_atom):
    return DeonticAtoms.OMISSIBLE.value.wrap(deontic_atom)


def permitted(deontic_atom):
    return DeonticAtoms.PERMITTED.value.wrap(deontic_atom)


def permitted_by_default(deontic_atom):
    return DeonticAtoms.PERMITTED_BY_DEFAULT.value.wrap(deontic_atom)


def omissible_by_default(deontic_atom):
    return DeonticAtoms.OMISSIBLE_BY_DEFAULT.value.wrap(deontic_atom)


def violated_obligation(deontic_atom):
    return DeonticAtoms.VIOLATED_OBLIGATION.value.wrap(deontic_atom)


def fulfilled_obligation(deontic_atom):
    return DeonticAtoms.FULFILLED_OBLIGATION.value.wrap(deontic_atom)


def non_violated_obligation(deontic_atom):
    return DeonticAtoms.NON_VIOLATED_OBLIGATION.value.wrap(deontic_atom)


def non_fulfilled_obligation(deontic_atom):
    return DeonticAtoms.NON_FULFILLED_OBLIGATION.value.wrap(deontic_atom)


def undetermined_obligation(deontic_atom):
    return DeonticAtoms.UNDETERMINED_OBLIGATION.value.wrap(deontic_atom)


def default_obligation(deontic_atom):
    return DeonticAtoms.DEFAULT_OBLIGATION.value.wrap(deontic_atom)


def violated_prohibition(deontic_atom):
    return DeonticAtoms.VIOLATED_PROHIBITION.value.wrap(deontic_atom)


def fulfilled_prohibition(deontic_atom):
    return DeonticAtoms.FULFILLED_PROHIBITION.value.wrap(deontic_atom)


def non_violated_prohibition(deontic_atom):
    return DeonticAtoms.NON_VIOLATED_PROHIBITION.value.wrap(deontic_atom)


def non_fulfilled_prohibition(deontic_atom):
    return DeonticAtoms.NON_FULFILLED_PROHIBITION.value.wrap(deontic_atom)


def undetermined_prohibition(deontic_atom):
    return DeonticAtoms.UNDETERMINED_PROHIBITION.value.wrap(deontic_atom)


def default_prohibition(deontic_atom):
    return DeonticAtoms.DEFAULT_PROHIBITION.value.wrap(deontic_atom)




import sys

import clingo
import telingo
import tempfile

from clingo import ast, SymbolType, Function

from deolingo._translator import DeolingoTranslator
from deolingo._deontic_atom import _DEOLINGO_ATOM_PREFIX


class DeolingoTelingoApp(telingo.TelApp):

    def run(self, program, files, weak=False):
        with ast.ProgramBuilder(program) as builder:
            transformer = DeolingoTranslator(builder.add, translate=True, temporal=True, weak=weak)
            transformer.transform_sources(None, files)
        translated_program = transformer.translated_program
        if "--translate" in sys.argv:
            print(translated_program)
            return

        # Create a named temporary file (automatically deleted on close)
        with tempfile.NamedTemporaryFile(mode="w+t", delete=False) as temp_file:
            temp_file.write(translated_program)
            temp_file.seek(0)  # Rewind to the beginning of the file
            clingo.clingo_main(self, [temp_file.name])
            temp_file.close()  # Important: close the file first to release the handle
            import os
            os.remove(temp_file.name)

    def print_model(self, model, printer):
        table = {}
        for sym in model.symbols(shown=True):
            if sym.type == SymbolType.Function and len(sym.arguments) > 0:
                table.setdefault(sym.arguments[-1].number, []).append(
                    Function(sym.name, sym.arguments[:-1], sym.positive))
        for step in range(self._TelApp__horizon + 1):
            symbols = table.get(step, [])
            sys.stdout.write(" State {}:".format(step))
            sig = None
            for sym in sorted(symbols):
                if not sym.name.startswith('__'):
                    import re
                    pattern = "^(-?)" + _DEOLINGO_ATOM_PREFIX + r"(\w+)\((.+)\)$"
                    # extract the first match group of the pattern from sym
                    m = re.match(pattern, str(sym))
                    if m is not None:
                        if m.group(2) in ["obligatory", "permitted", "forbidden", "omissible"]\
                                and m.group(1) != "-" and not m.group(3).startswith('-'):
                            sym = re.sub(pattern, r"\1&\2{\3}", str(sym))
                            sys.stdout.write("\n ")
                            sys.stdout.write(" {}".format(sym))
                        if m.group(2) == "inconsistency":
                            sym = re.sub(pattern, _DEOLINGO_ATOM_PREFIX + r"\2(\3)", str(sym))
                            sys.stdout.write("\n ")
                            sys.stdout.write(" {}".format(sym))
                    else:
                        if (sym.name, len(sym.arguments), sym.positive) != sig:
                            sys.stdout.write("\n ")
                            sig = (sym.name, len(sym.arguments), sym.positive)
                        sys.stdout.write(" {}".format(sym))
            sys.stdout.write("\n")
        return True




import copy

from clingo import ast as ast
from clingo.ast import Transformer
from clingox.ast import theory_term_to_term, theory_term_to_literal

from deolingo._deontic_atom import DeonticAtoms


class DeonticASTTransformer(Transformer):

    # <editor-fold desc="Initialization">

    def __init__(self, translate=False):
        super().__init__()
        self.deontic_atoms = set()
        self.translate = translate
        self.translated_program = ""
        self._last_literal_sign = None
        self._head_theory_atoms_sequence = []
        self._body_theory_atoms_sequence = []
        self._in_body = False
        self._in_head = False
        self._in_rule = False
        self._deontic_conditional = None
        self._file = None

    # </editor-fold>

    # <editor-fold desc="clingo.ast.Transformer override">

    def visit_Comment(self, comment):
        if not comment.value.startswith("%!"):
            raise MultipleRuleException([])
        self._add_to_translation(comment.location, comment)
        return comment

    def visit_Rule(self, rule):
        self._in_rule = True
        self._in_head = True
        self._head_theory_atoms_sequence = []
        self._body_theory_atoms_sequence = []
        self._deontic_conditional = None
        self._last_literal_sign = None
        multi_rule = None
        new_head = self.visit(rule.head)
        multi_rule, new_head = self._process_theory_atoms_in_head(multi_rule, new_head, rule)
        self._in_head = False
        self._in_body = True
        new_body = self.visit_sequence(rule.body)
        new_body = self._filter_out_theory_atoms_and_literals_of_theory_atoms(new_body)
        new_body.extend(self._body_theory_atoms_sequence)
        self._in_body = False
        assert new_head is not None
        assert new_body is not None
        if multi_rule is not None:
            for rule in multi_rule:
                rule.body.extend(new_body)
            raise MultipleRuleException(multi_rule, rule.location)
        new_rule = ast.Rule(rule.location, new_head, new_body)
        self._add_to_translation(rule.location, new_rule)
        return new_rule

    def visit_Literal(self, literal):
        self._last_literal_sign = None
        if literal.atom.ast_type == ast.ASTType.TheoryAtom:
            self._last_literal_sign = literal.sign
        self.visit(literal.atom)
        return literal

    def visit_TheoryAtom(self, atom):
        if atom.term.name == "tel":
            return atom
        if atom.term.name == "show":
            self._map_show_atom(atom)

        def has_valid_condition(lit):
            return 'condition' in lit and lit['condition'] is not None and \
                lit['condition'] != []
        if self._in_head:
            iterable, literals, deontic_conditional = self._map_deontic_atom_with_sequence(atom)
            if deontic_conditional is not None:
                self._deontic_conditional = deontic_conditional
            if not iterable:
                return atom
            conditional_literals = [ast.ConditionalLiteral(atom.term.location, lit['term'], lit['condition'] or [])
                                    for lit in literals]
            self._head_theory_atoms_sequence.extend(conditional_literals)
        elif self._in_body:
            iterable, literals, deontic_conditional = self._map_deontic_atom_with_sequence(atom)
            if deontic_conditional is not None:
                raise Exception(f"Deontic conditional '{deontic_conditional}' not allowed in body")
            if not iterable:
                return atom
            conditional_literals = [ast.ConditionalLiteral(atom.term.location, lit['term'], lit['condition'])
                                    if has_valid_condition(lit) else lit['term']
                                    for lit in literals]
            self._body_theory_atoms_sequence.extend(conditional_literals)
        return atom

    # </editor-fold>

    # <editor-fold desc="Private methods">

    def _is_telingo_atom(self, atom):
        return (atom.ast_type == ast.ASTType.TheoryAtom and \
                hasattr(atom, "term") and atom.term.name == "tel")


    def _process_theory_atoms_in_head(self, multi_rule, new_head, rule):
        if new_head.ast_type == ast.ASTType.Disjunction or new_head.ast_type == ast.ASTType.TheoryAtom and \
               not self._is_telingo_atom(new_head):
            new_head_elements = self._filter_out_theory_atoms_and_literals_of_theory_atoms(
                new_head.elements) if new_head.ast_type == ast.ASTType.Disjunction else []
            if self._deontic_conditional is not None:
                new_rule = ast.Rule(rule.location,
                                    self._deontic_conditional.term,
                                    [self._deontic_conditional.condition])
                new_rule_2 = copy.deepcopy(new_rule)
                ob_nv_atom_name = DeonticAtoms.NON_VIOLATED_OBLIGATION.value.prefixed()
                ob_nv_term = self._deontic_conditional.condition.atom.symbol
                ob_nv_atom = _symbolic_atom(rule.location, ob_nv_atom_name, [ob_nv_term])
                new_rule_2.body = [_positive_literal(rule.location, ob_nv_atom)]
                multi_rule = [new_rule, new_rule_2]
            else:
                new_head_elements.extend(self._head_theory_atoms_sequence)
                new_head = ast.Disjunction(new_head.location, new_head_elements)
        return multi_rule, new_head

    def _add_to_translation(self, location, statement):
        if self.translate:
            if location is not None:
                if self._file != location.end.filename:
                    self._file = location.end.filename
                    self.translated_program += f"\n% File: {location.end.filename}\n"
                self.translated_program += f"\n% Source line: {location.begin.line}\n"
            self.translated_program += str(statement) + "\n"

    def _map_deontic_atom_with_sequence(self, atom):
        if len(atom.elements) < 1:
            raise Exception(f"Empty deontic atom '{atom}'")
        deontic_atom = DeonticAtoms.with_name(atom.term.name)
        if deontic_atom is None:
            if atom.term.name == "tel":
                return True, [atom], None
            raise Exception(f"Deontic atom '{atom.term}' not found")
        deontic_conditional = _DeonticConditional.from_deontic_atom(atom)
        if deontic_conditional is not None:
            if deontic_conditional.symbolic_term.ast_type != ast.ASTType.Variable:
                dt_name = str(deontic_conditional.symbolic_term)
                dt_is_negated = dt_name.startswith("-")
                dt_name = dt_name[1:] if dt_is_negated else dt_name
                self.deontic_atoms.add(dt_name)
            return False, None, deontic_conditional
        new_name = deontic_atom.prefixed()
        sign = self._last_literal_sign if self._last_literal_sign is not None else ast.Sign.NoSign
        new_deontic_atoms = []

        def symbolize_terms(terms):
            new_terms = [theory_term_to_term(t) for t in terms]
            for t in new_terms:
                if t.ast_type != ast.ASTType.Variable:
                    new_deontic_atoms.append(t)
            return _symbolic_literal(atom.term.location, sign, new_name, new_terms)
        is_disjunction = _DeonticDisjunctionOrConjunction.transform_deontic_atom(atom, operator="||")
        if is_disjunction and self._in_body:
            raise Exception(f"Disjunction '{atom}' not allowed in body")
        is_conjunction = _DeonticDisjunctionOrConjunction.transform_deontic_atom(atom, operator="&&")
        if is_conjunction and self._in_head:
            raise Exception(f"Conjunction '{atom}' not allowed in head")
        new_atoms = [{'term': symbolize_terms(the_term.terms), 'condition': the_term.condition}
                     for the_term in atom.elements]
        for new_deontic_atom in new_deontic_atoms:
            if new_deontic_atom.ast_type != ast.ASTType.Variable:
                dt_name = str(new_deontic_atom)
                dt_is_negated = dt_name.startswith("-")
                dt_name = dt_name[1:] if dt_is_negated else dt_name
                self.deontic_atoms.add(dt_name)
        return True, new_atoms, None

    def _map_show_atom(self, atom):
        show_atoms = set()
        for element in atom.elements:
            deontic_atom = DeonticAtoms.with_name(str(element.terms[0]))
            if deontic_atom is not None:
                show_atoms.add(deontic_atom.prefixed())
        show_rules = []
        for sa in show_atoms:
            show_rules.append(ast.ShowSignature(atom.location, sa, 1, 1))
        raise MultipleRuleException(show_rules, atom.location)

    def _filter_out_theory_atoms_and_literals_of_theory_atoms(self, elements):
        return [element for element in elements
                if (element.ast_type != ast.ASTType.TheoryAtom and
                (element.ast_type != ast.ASTType.Literal or
                 element.atom.ast_type != ast.ASTType.TheoryAtom)) or
                (element.ast_type == ast.ASTType.TheoryAtom and element.term.name == "tel") or
                (element.ast_type == ast.ASTType.Literal and
                 element.atom.ast_type == ast.ASTType.TheoryAtom and
                 element.atom.term.name == "tel")
                ]

    # </editor-fold>


# <editor-fold desc="Exception classes">

class MultipleRuleException(Exception):
    def __init__(self, rules, line=None):
        self.rules = rules
        self.line = line

# </editor-fold>


# <editor-fold desc="Private classes">

class _DeonticConditional:

    def __init__(self, deontic_atom, term, condition, symbolic_term, string_repr=None):
        self.deontic_atom = deontic_atom
        self.term = term
        self.condition = condition
        self.string_repr = string_repr
        self.symbolic_term = symbolic_term

    @staticmethod
    def from_deontic_atom(atom):
        if len(atom.elements) < 1:
            return None
        deontic_atom = DeonticAtoms.with_name(atom.term.name)
        if deontic_atom is None:
            return None
        deontic_atom_prefixed = deontic_atom.prefixed()
        for element in atom.elements:
            is_conditional = len(element.terms) == 1 and \
                             element.terms[0].ast_type == ast.ASTType.TheoryUnparsedTerm and \
                             len(element.terms[0].elements) == 2 and \
                             element.terms[0].elements[0].ast_type == ast.ASTType.TheoryUnparsedTermElement and \
                             element.terms[0].elements[1].ast_type == ast.ASTType.TheoryUnparsedTermElement and \
                             element.terms[0].elements[1].operators[0] == "|"
            if is_conditional:
                lhs_operators = element.terms[0].elements[0].operators
                rhs_operators = element.terms[0].elements[1].operators
                lhs_negative = lhs_operators[0] == "-" if len(lhs_operators) > 0 else False
                rhs_negative = rhs_operators[1] == "-" if len(rhs_operators) > 1 else False
                symbolic_term = theory_term_to_term(element.terms[0].elements[0].term)
                if lhs_negative:
                    symbolic_term = ast.UnaryOperation(atom.location, ast.UnaryOperator.Minus, symbolic_term)
                term = _symbolic_literal(atom.location, ast.Sign.NoSign, deontic_atom_prefixed, [symbolic_term])
                condition = theory_term_to_literal(element.terms[0].elements[1].term)
                if rhs_negative:
                    rhs_term = theory_term_to_term(element.terms[0].elements[1].term)
                    rhs_atom = ast.SymbolicAtom(ast.UnaryOperation(atom.location, ast.UnaryOperator.Minus, rhs_term))
                    condition = _positive_literal(atom.location, rhs_atom)
                return _DeonticConditional(deontic_atom_prefixed, term, condition, symbolic_term, str(atom))
        return None

    def __str__(self):
        return self.string_repr


class _DeonticDisjunctionOrConjunction:

    @staticmethod
    def transform_deontic_atom(atom, operator="||"):
        if len(atom.elements) < 1:
            return None
        deontic_atom = DeonticAtoms.with_name(atom.term.name)
        if deontic_atom is None:
            return None
        new_elements = []
        contains_operation = False
        for element in atom.elements:
            is_operation = False
            for term in element.terms:
                if term.ast_type == ast.ASTType.TheoryUnparsedTerm and len(term.elements) > 1:
                    pass
                else:
                    continue
                is_operation = term.elements[0].ast_type == ast.ASTType.TheoryUnparsedTermElement and \
                               all(x.ast_type == ast.ASTType.TheoryUnparsedTermElement and x.operators[0] == operator
                                   for x in term.elements[1:])
                if is_operation:
                    for el in term.elements:
                        new_element_term = theory_term_to_term(el.term)
                        new_element = ast.TheoryAtomElement([new_element_term], [])
                        new_elements.append(new_element)
            contains_operation = contains_operation or is_operation
            if not is_operation:
                new_elements.append(element)
        atom.elements = new_elements
        return contains_operation

# </editor-fold>


# <editor-fold desc="Helper functions">

def _positive_literal(location, atom):
    return ast.Literal(location, ast.Sign.NoSign, atom)


def _negative_literal(location, atom):
    return ast.Literal(location, ast.Sign.Negation, atom)


def _symbolic_atom(location, name, terms):
    return ast.SymbolicAtom(ast.Function(location, name, terms, False))


def _symbolic_literal(location, sign, atom_name, terms):
    return ast.Literal(location, sign, _symbolic_atom(location, atom_name, terms))

# </editor-fold>




from enum import Enum

from langchain_community.llms import GPT4All
from langchain_openai.llms import OpenAI
from langchain_community.llms import HuggingFaceEndpoint
from langchain.prompts import PromptTemplate
import google.generativeai as genai
import os


class Generator(str, Enum):
    GEMINI = "gemini"
    GPT4ALL = "gpt4all"
    OPENAI = "openai"
    HUGGINGFACE = "hugging"


class DeonticProgramGenerator:

    def __init__(self, generator=Generator.GPT4ALL, llm=None):
        self._generator = generator
        self._llm = llm
        self._openai_api_key = os.getenv('OPENAI_API_KEY', None)
        self._gemini_api_key = os.getenv('GEMINI_API_KEY', None)

    def generate_program(self, natural_language_text):
        if self._generator == Generator.GEMINI and self._gemini_api_key is not None:
            print("Using Gemini for generation")
            return self._generate_program_gemini(natural_language_text, model_name=self._llm)
        else:
            return self._generate_program_langchain(natural_language_text, model=self._llm, model_path=self._llm)

    def _generate_program_langchain(self, natural_language_text, model="meta-llama/CodeLlama-70b-Instruct-hf",
                                    model_path="./deolingo/models/mistral-7b-instruct-v0.1.Q4_0.gguf"):
        if self._generator == Generator.OPENAI and self._openai_api_key is not None:
            print("Using OpenAI for generation")
            llm = OpenAI(model_name="gpt-3.5-turbo-16k", temperature=0, openai_api_key=self._openai_api_key)
        elif self._generator == Generator.HUGGINGFACE:
            print("Using HuggingFaceHub for generation")
            llm = HuggingFaceEndpoint(repo_id=model, timeout=240)
        else:
            print(f"Using local GPT4All model for generation: {model_path}")
            llm = GPT4All(model=model_path)
        prompt = PromptTemplate(input_variables=["preprompt", "text"], template=_long_prompt_template)
        chain = prompt | llm
        output = chain.invoke({
            "preprompt": _pre_prompt,
            "text": natural_language_text
        })
        return output

    def _generate_program_gemini(self, natural_language_text, model_name="gemini-1.5-flash-latest"):
        genai.configure(api_key=self._gemini_api_key)
        model = genai.GenerativeModel(model_name=model_name)
        response = model.generate_content([_short_prompt_template, natural_language_text])
        return response.candidates[0].content.parts[0].text


_long_prompt_template = """
{preprompt}

Write a deontic logic program in Clingo based on the following description:
{text}
"""

_short_prompt_template = """
You are a Clingo expert and a Deontic logic expert. 
Using the following theory atoms &obligatory{{}} and &forbidden{{}}.
Write a deontic logic program in Clingo based on the following description:
{text}
"""

_pre_prompt = """
Consider this docs and examples:

Deolingo is a tool for reasoning about Deontic Logic based in Answer Set Programming.
The tool is implemented in Python and uses [Clingo](https://potassco.org/clingo) as the ASP solver.

## Examples

Simple example:
```
    % Example: It is permitted to park if there is no evidence that it is forbidden to park.
    &permitted{park} :- not &forbidden{park}.
```n

You can find examples of Deontic Equilibrium Logic programs in the [examples](deolingo/examples) directory.
See example from [example 2](deolingo/examples/preliminary/example2.lp)
```
% I must normally work (it is obligatory that I work if it is not explicitly omissible).
&obligatory{work} :- not &omissible{work}.

% On weekends, I have an explicit permission not to work.
&omissible{work} :- weekend.

% It is not a weekend.
-weekend.

% I decided not to work.
-work.
````

Output from deolingo:

```
Answer: 1
FACTS: -work, -weekend
OBLIGATIONS: &obligatory{work}
PROHIBITIONS:
SATISFIABLE
```

The output shows that the program is satisfiable and there is an answer set whose atoms mean:

- FACTS:
  - `-work`: I did not work.
  - `-weekend`: It is not a weekend.
- OBLIGATIONS:
- `&obligatory{work}`: It is obligatory that I work.
  The obligation to work is violated, although that is not checked in any rule and it is excluded from the answer set.
- PROHIBITIONS:
  - Nothing remarkable in the _world_ of prohibitions.


```

## Deontic theory atoms

Deolingo uses the following atoms to represent the deontic theory:

### Main atoms

- `&obligatory{p}`: It is **obligatory** that `p`. Equivalent to forbidden not `p`: `&forbidden{-p}`. Also known as mandatory, duty and required.
- `&forbidden{p}`: It is **forbidden** that `p`. Equivalent to obligatory not `p`: `&obligatory{-p}`. Also known as prohibited and impermissible.
- `&omissible{p}`: It is **omissible** that `p`. Equivalent to not explicitly obligatory: `-&obligatory{p}` (but this syntax is not allowed, see ["Syntax limitations"](#syntax-limitations) below).
- `&permitted{p}`: It is **permitted** (or permissible) that `p`. Equivalent to not explicitly forbidden: `-&obligatory{p}` (but this syntax is not allowed, see ["Syntax limitations"](#syntax-limitations) below).
- `&optional{p}`: It is **optional** that `p`. Equivalent to omissible and permissible. Cannot be neither obligatory nor forbidden.
- `&deontic{p}`: `p` is a **deontic** atom. It can be obligatory, forbidden, omissible, permitted or optional.

Deolingo also reifies the truth of the deontic atoms to allow general rules requiring it.
It is represented by the _holds_ theory atom `&holds{p}`, which is true if, and only if, the atom `p` is true.

### Violation and fulfillment atoms

- `&violated_obligation{p}`: The obligation of `p` is violated.
- `&violated_prohibition{p}`: The obligation of `p` is fulfilled.
- `&non_fulfilled_obligation{p}`: The obligation of `p` is not fulfilled.
- `&non_violated_obligation{p}`: The obligation of `p` is not violated.
- `&undetermined_obligation{p}`: The obligation of `p` is undetermined. It is neither violated nor fulfilled.
- `&violated_prohibition{p}`: The prohibition of `p` is violated.
- `&fulfilled_prohibition{p}`: The prohibition of `p` is fulfilled.
- `&non_fulfilled_prohibition{p}`: The prohibition of `p` is not fulfilled.
- `&non_violated_prohibition{p}`: The prohibition of `p` is not violated.
- `&undetermined_prohibition{p}`: The prohibition of `p` is undetermined. It is neither violated nor fulfilled.


### Implicit atoms and defaults

- `&default_obligation{p}`: It is obligatory by **default** that `p`.
  If there is no evidence of not `p` being permitted, it is obligatory by default that `p`.
- `&default_prohibition{p}`: It is forbidden by **default** that `p`.
  If there is no evidence of not `p` being omissible, it is forbidden by default that `p`.
- `&om_d{p}`: It is omissible by **default** that `p`.
  If there is no evidence of `p` being obligatory, it is omissible by default that `p`.
- `&permitted_by_default{p}`: It is permitted by **default** that `p`.
  If there is no evidence of `p` being forbidden, it is permitted by default that `p`.
- `&omissible_implicitly{p}`: It is omissible **implicitly** that `p`. Requires default omissibility.
  Equivalent to not obligatory (default negation): there is no evidence of obligation: `not &obligatory{p}.`
- `&permitted_implicitly{p}`: It is permitted **implicitly** that `p`. Requires default permissibility.
  Equivalent to not obligatory (default negation): there is no evidence of prohibition: `not &forbidden{p}.`.

### Clingo theory definition for Deolingo
```
#theory deolingo {
    deontic_term {
        - : 4, unary;
        && : 3, binary, left;
        || : 2, binary, left;
        | : 1, binary, left
    };
    show_term { / : 1, binary, left };
    &obligatory/0 : deontic_term, any;
    &forbidden/0 : deontic_term, any;
    &omissible/0 : deontic_term, any;
    &permitted/0 : deontic_term, any;
    &optional/0 : deontic_term, any;
    &permitted_by_default/0 : deontic_term, any;
    &omissible_by_default/0 : deontic_term, any;
    &holds/0 : deontic_term, any;
    &deontic/0 : deontic_term, any;
    &permitted_implicitly/0 : deontic_term, any;
    &omissible_implicitly/0 : deontic_term, any;
    &violated/0 : deontic_term, any;
    &fulfilled/0 : deontic_term, any;
    &violated_obligation/0 : deontic_term, any;
    &fulfilled_obligation/0 : deontic_term, any;
    &non_violated_obligation/0 : deontic_term, any;
    &non_fulfilled_obligation/0 : deontic_term, any;
    &undetermined_obligation/0 : deontic_term, any;
    &default_obligation/0 : deontic_term, any;
    &violated_prohibition/0 : deontic_term, any;
    &fulfilled_prohibition/0 : deontic_term, any;
    &non_violated_prohibition/0 : deontic_term, any;
    &non_fulfilled_prohibition/0 : deontic_term, any;
    &undetermined_prohibition/0 : deontic_term, any;
    &default_prohibition/0 : deontic_term, any;
    &show/0 : show_term, directive
}.
```

### Deontic theory atoms syntax

Currently, Deolingo supports the following constructs inside the deontic theory atoms:
- Simple terms and their explicit negations. 
  For example, `&obligatory{p}.` and `&obligatory{-p}.`.
- Variables. For example, `&obligatory{X}.` and `&obligatory{-X}.`.
- Conditional literals: `:- &obligatory{p: q}.` is equivalent to `:- &obligatory{p} : q.`.
- Sequences of simple terms and variables with conditional literals.
  For example, `&obligatory{p; X}.`, meaning `&obligatory{p}; &obligatory{X}.`.
- Disjunctions in the **head** of rules: `&obligatory{p || q}.`, which is equivalent to `&obligatory{p; q}.`.
- Conjunctions in the **body** of rules: `:- &obligatory{p && q}.`, which is equivalent to `:- &obligatory{p; q}.`.
- Deontic conditional operator (in the **head** of rules): `&obligatory{p | q}.`, 
  meaning `&obligatory{p} :- q. &obligatory{p} :- &non_violated_obligation{q}.`, as described in the DELX paper.
"""





from deolingo._rewriting_translator import DeolingoRewritingTranslator
from deolingo._translator import DeolingoTranslator
from deolingo.domain.use_case_command import UseCaseCommand

import clingo.ast as ast


class SolveDeonticProgramCommand(UseCaseCommand):

    def __init__(self, program, files, translate=False, optimize=False, weak=False):
        super().__init__()
        self.program = program
        self.files = files
        self.translate = translate
        self.optimize = optimize
        self.weak = weak

    def execute(self):
        with ast.ProgramBuilder(self.program) as builder:
            if self.optimize:
                transformer = DeolingoRewritingTranslator(builder.add, translate=self.translate, weak=self.weak)
            else:
                transformer = DeolingoTranslator(builder.add, translate=self.translate, weak=self.weak)
            transformer.transform_sources(None, self.files)
        if self.translate:
            print(transformer.translated_program)
            return
        self.program.ground([("base", [])])
        self.program.solve(on_model=None, async_=False)




import clingo
import sys

from dotenv import load_dotenv

from deolingo._application import DeolingoApplication


def main():
    """Run the deolingo application."""
    load_dotenv()
    app = DeolingoApplication()
    sys.exit(int(clingo.clingo_main(app)))


if __name__ == '__main__':
    main()




from deolingo._generator import DeonticProgramGenerator
from deolingo.domain.use_case_command import UseCaseCommand


class GenerateDeonticProgramCommand(UseCaseCommand):

    def __init__(self, input_text, generator_type=None, llm=None):
        super().__init__()
        self.input_text = input_text
        self.generator = DeonticProgramGenerator(generator=generator_type, llm=llm)

    def execute(self, params=None):
        program = self.generator.generate_program(self.input_text)
        print(program)





from deolingo.domain.use_case_command import UseCaseCommand
from deolingo.xcontrol import XDeolingoControl


class ExplainDeonticProgramCommand(UseCaseCommand):

    def __init__(self, inputs, n_solutions='1', _n_explanations='1', translate=False, weak=False):
        super().__init__()
        self.inputs = inputs
        self.translate = translate
        self._xcontrol = XDeolingoControl(n_solutions=n_solutions, n_explanations=_n_explanations, weak=weak)

    def execute(self):
        self._xcontrol.add_inputs(self.inputs)
        if self.translate:
            print(self._xcontrol.rewritten_program)
            return
        self._xcontrol.ground([("base", [])])
        self._print_text_explanations()

    def _print_text_explanations(self):
        n = 0
        for answer in self._xcontrol.explain():
            n += 1
            print(f'Answer {n}')
            for expl in answer:
                print(expl.ascii_tree())



import unittest

from deolingo.examples import ExampleReader
from deolingo.solver import DeolingoSolver


class DeolingoRegressionTestCase(unittest.TestCase):

    @staticmethod
    def get_sut(all_models=True):
        return DeolingoSolver(all_models=all_models)

    def test_issue_10_error_1(self):
        """Test for the 1st error on issue https://github.com/ovidiomanteiga/deolingo/issues/10"""
        # Arrange
        example = """
            &obligatory{c} :- not &obligatory{m}.
            &obligatory{s}.
        """
        expected_answer_set = {'&obligatory{s}', '&obligatory{c}'}
        # Act
        actual_answer_sets = self.get_sut().solve(example)
        # Assert
        assert len(actual_answer_sets) == 1
        actual_answer_set = set(actual_answer_sets[0])
        assert actual_answer_set == expected_answer_set

    def test_issue_10_error_2(self):
        """Test for 2nd error on issue https://github.com/ovidiomanteiga/deolingo/issues/10"""
        # Arrange
        example = """
            &obligatory{c} :- not &obligatory{m}.
            &obligatory{m} :- not &obligatory{c}.
        """
        expected_answer_set_0 = {'&obligatory{m}'}
        expected_answer_set_1 = {'&obligatory{c}'}
        # Act
        actual_answer_sets = self.get_sut().solve(example)
        print(actual_answer_sets)
        # Assert
        assert len(actual_answer_sets) == 2
        assert set(actual_answer_sets[0]) == expected_answer_set_0
        assert set(actual_answer_sets[1]) == expected_answer_set_1


class DeolingoRegressionOptimizedTestCase(DeolingoRegressionTestCase):

    @staticmethod
    def get_sut(all_models=True):
        return DeolingoSolver(all_models=all_models, optimize=True)




import dataclasses
from os import listdir
from os.path import isfile, join, isdir
from pathlib import Path


class ExampleReader:

    def __init__(self):
        self._examples_folder = Path(__file__).parent / "examples"

    def read_examples(self):
        """Returns a list of all examples reading from folders in /examples"""
        folders = [node for node in listdir(self._examples_folder)
                   if isdir(join(self._examples_folder, node))]
        examples_files = []
        for folder in folders:
            folder = self._examples_folder / folder
            examples_files += [folder / file for file in listdir(folder)
                               if isfile(join(folder, file)) and file.endswith(".lp")]
        files = [(self._name(file), open(file)) for file in examples_files]
        examples = [DeolingoExample(name=file[0], contents=file[1].read()) for file in files]
        return examples

    def read_example(self, example):
        """Returns an example from the examples in the /examples folder."""
        if not example.endswith(".lp") or not isfile(join(self._examples_folder, example)):
            return None
        example_file = self._examples_folder / example
        file_name = self._name(example_file)
        file_contents = open(example_file).read()
        example = DeolingoExample(name=file_name, contents=file_contents)
        return example

    @staticmethod
    def _name(file):
        return file.parent.name + "/" + file.name


@dataclasses.dataclass
class DeolingoExample:
    name: str
    contents: str




from deolingo.xcontrol import XDeolingoControl


class XDeolingoSolver:

    def __init__(self, xcontrol=None):
        self._xcontrol = xcontrol if xcontrol is not None else XDeolingoControl()

    def solve(self, inputs):
        self._xcontrol = XDeolingoControl(n_solutions='0', n_explanations='0', auto_trace='none')
        self._xcontrol.add_inputs(inputs)
        self._xcontrol.ground([("base", [])])
        return self._xcontrol.explain()




from deolingo._deontic_atom import DeonticAtoms as Deo


_DEOLINGO_THEORY = f"""
#theory deolingo {{
    deontic_term {{
        - : 4, unary;
        && : 3, binary, left;
        || : 2, binary, left;
        | : 1, binary, left
    }};
    show_term {{ / : 1, binary, left }};
    &{Deo.OBLIGATORY.value.name}/0 : deontic_term, any;
    &{Deo.FORBIDDEN.value.name}/0 : deontic_term, any;
    &{Deo.OMISSIBLE.value.name}/0 : deontic_term, any;
    &{Deo.PERMITTED.value.name}/0 : deontic_term, any;
    &{Deo.OPTIONAL.value.name}/0 : deontic_term, any;
    &{Deo.PERMITTED_BY_DEFAULT.value.name}/0 : deontic_term, any;
    &{Deo.OMISSIBLE_BY_DEFAULT.value.name}/0 : deontic_term, any;
    &{Deo.HOLDS.value.name}/0 : deontic_term, any;
    &{Deo.DEONTIC.value.name}/0 : deontic_term, any;
    &{Deo.PERMITTED_IMPLICITLY.value.name}/0 : deontic_term, any;
    &{Deo.OMISSIBLE_IMPLICITLY.value.name}/0 : deontic_term, any;
    &{Deo.VIOLATED.value.name}/0 : deontic_term, any;
    &{Deo.FULFILLED.value.name}/0 : deontic_term, any;
    &{Deo.VIOLATED_OBLIGATION.value.name}/0 : deontic_term, any;
    &{Deo.FULFILLED_OBLIGATION.value.name}/0 : deontic_term, any;
    &{Deo.NON_VIOLATED_OBLIGATION.value.name}/0 : deontic_term, any;
    &{Deo.NON_FULFILLED_OBLIGATION.value.name}/0 : deontic_term, any;
    &{Deo.UNDETERMINED_OBLIGATION.value.name}/0 : deontic_term, any;
    &{Deo.DEFAULT_OBLIGATION.value.name}/0 : deontic_term, any;
    &{Deo.VIOLATED_PROHIBITION.value.name}/0 : deontic_term, any;
    &{Deo.FULFILLED_PROHIBITION.value.name}/0 : deontic_term, any;
    &{Deo.NON_VIOLATED_PROHIBITION.value.name}/0 : deontic_term, any;
    &{Deo.NON_FULFILLED_PROHIBITION.value.name}/0 : deontic_term, any;
    &{Deo.UNDETERMINED_PROHIBITION.value.name}/0 : deontic_term, any;
    &{Deo.DEFAULT_PROHIBITION.value.name}/0 : deontic_term, any;
    &show/0 : show_term, directive
}}.
"""


_DEOLINGO_RESTRICTED_THEORY = f"""
#theory deolingo_restricted {{
    deontic_term {{
        - : 4, unary;
        && : 3, binary, left;
        || : 2, binary, left;
        | : 1, binary, left
    }};
    show_term {{ / : 1, binary, left }};
    &{Deo.OBLIGATORY.value.name}/0 : deontic_term, any;
    &{Deo.FORBIDDEN.value.name}/0 : deontic_term, any;
    &{Deo.OMISSIBLE.value.name}/0 : deontic_term, any;
    &{Deo.PERMITTED.value.name}/0 : deontic_term, any;
    &{Deo.OPTIONAL.value.name}/0 : deontic_term, any;
    &{Deo.PERMITTED_BY_DEFAULT.value.name}/0 : deontic_term, head;
    &{Deo.OMISSIBLE_BY_DEFAULT.value.name}/0 : deontic_term, head;
    &{Deo.HOLDS.value.name}/0 : deontic_term, any;
    &{Deo.DEONTIC.value.name}/0 : deontic_term, any;
    &{Deo.PERMITTED_IMPLICITLY.value.name}/0 : deontic_term, body;
    &{Deo.OMISSIBLE_IMPLICITLY.value.name}/0 : deontic_term, body;
    &{Deo.VIOLATED.value.name}/0 : deontic_term, body;
    &{Deo.FULFILLED.value.name}/0 : deontic_term, body;
    &{Deo.VIOLATED_OBLIGATION.value.name}/0 : deontic_term, body;
    &{Deo.FULFILLED_OBLIGATION.value.name}/0 : deontic_term, body;
    &{Deo.NON_VIOLATED_OBLIGATION.value.name}/0 : deontic_term, body;
    &{Deo.NON_FULFILLED_OBLIGATION.value.name}/0 : deontic_term, body;
    &{Deo.UNDETERMINED_OBLIGATION.value.name}/0 : deontic_term, body;
    &{Deo.DEFAULT_OBLIGATION.value.name}/0 : deontic_term, head;
    &{Deo.VIOLATED_PROHIBITION.value.name}/0 : deontic_term, body;
    &{Deo.FULFILLED_PROHIBITION.value.name}/0 : deontic_term, body;
    &{Deo.NON_VIOLATED_PROHIBITION.value.name}/0 : deontic_term, body;
    &{Deo.NON_FULFILLED_PROHIBITION.value.name}/0 : deontic_term, body;
    &{Deo.UNDETERMINED_PROHIBITION.value.name}/0 : deontic_term, body;
    &{Deo.DEFAULT_PROHIBITION.value.name}/0 : deontic_term, head;
    &show/0 : show_term, directive
}}.
"""




from deolingo._benchmark import BenchmarkRunner
from deolingo.domain.use_case_command import UseCaseCommand


class RunBenchmarkCommand(UseCaseCommand):

    def __init__(self):
        super().__init__()
        self.runner = BenchmarkRunner()

    def execute(self):
        self.runner.run_benchmark()
        self.runner.print_results()




from typing import Sequence, Optional, overload, Union, Tuple, Callable

from clingo import Control, Logger, ast, Symbol, Model, StatisticsMap, SolveResult, SolveHandle

from deolingo._answer_set_rewriter import DeonticAnswerSetRewriter
from deolingo._rewriting_translator import DeolingoRewritingTranslator
from deolingo._translator import DeolingoTranslator


class DeolingoControl(Control):
    """Extends the clingo.Control class to translate and solve deontic logic programs."""

    # <editor-fold desc="Initialization">

    def __init__(self, arguments: Sequence[str] = None, logger: Optional[Logger] = None,
                 message_limit: int = 20, grouped=False, optimize=False):
        if arguments is None:
            arguments = []
        super().__init__(arguments, logger, message_limit)
        self._translate_control = Control(arguments, logger, message_limit)
        self._program_builder = ast.ProgramBuilder(self._translate_control)
        self._transformer = DeolingoRewritingTranslator(self._program_builder.add, translate=True, add_theory=False) if optimize \
            else DeolingoTranslator(self._program_builder.add, translate=True)
        self._rewriter = DeonticAnswerSetRewriter(grouped=grouped)
        self._add_deolingo_theory_and_deontic_rules()

    # </editor-fold>

    # <editor-fold desc="clingo.Control override">

    def _add2(self, name: str, parameters: Sequence[str], program: str) -> None:
        translated = self._transformer.transform_source(program)
        super()._add2(name, parameters, translated)

    def solve(
        self,
        assumptions: Sequence[Union[Tuple[Symbol, bool], int]] = [],
        on_model: Optional[Callable[[Model], Optional[bool]]] = None,
        on_unsat: Optional[Callable[[Sequence[int]], None]] = None,
        on_statistics: Optional[Callable[[StatisticsMap, StatisticsMap], None]] = None,
        on_finish: Optional[Callable[[SolveResult], None]] = None,
        on_core: Optional[Callable[[Sequence[int]], None]] = None,
        yield_: bool = False,
        async_: bool = False,
    ) -> Union[SolveHandle, SolveResult]:
        return super().solve(
            assumptions=assumptions,
            on_model=self._on_model(on_model),
            on_unsat=on_unsat,
            on_statistics=on_statistics,
            on_finish=on_finish,
            on_core=on_core,
            yield_=yield_,
            async_=async_
        )

    # </editor-fold>

    # <editor-fold desc="Private methods">

    def _add_deolingo_theory_and_deontic_rules(self):
        translated = self._transformer.transform_sources("")
        super()._add2("base", [], translated)

    def _on_model(self, on_model_callback):
        """On model callback wrapper that rewrites the atoms of the model."""
        if on_model_callback is None:
            return None
        return lambda model: on_model_callback(self._rewriter.rewrite_model(model))

    # </editor-fold>




from typing import Sequence, Optional

from clingo import Logger

from deolingo._rewriting_translator import DeolingoRewritingTranslator
from deolingo.control import DeolingoControl


class DeolingoRewritingControl(DeolingoControl):
    """Extends the DeolingoControl class to use an optimized translator that rewrites the deontic atoms
     to their equivalents before solving the deontic logic program."""

    # <editor-fold desc="Initialization">

    def __init__(self, arguments: Sequence[str] = None, logger: Optional[Logger] = None,
                 message_limit: int = 20, grouped=False):
        super().__init__(arguments, logger, message_limit, grouped, optimize=True)

    # </editor-fold>




import copy

from clingo import ast as ast

from deolingo._ast_transformer import DeonticASTTransformer, _symbolic_atom, _positive_literal, \
    _negative_literal
from deolingo._deontic_atom import DeonticAtoms


class DeonticASTRewritingTransformer(DeonticASTTransformer):

    # <editor-fold desc="Initialization">

    def __init__(self, translate=False):
        super().__init__(translate=translate)
        self._derivation_rules_to_be_added = set()

    # </editor-fold>

    # <editor-fold desc="Getters/Setters">

    @property
    def derivation_rules_to_be_added(self):
        return self._derivation_rules_to_be_added

    # </editor-fold>

    # <editor-fold desc="clingo.ast.Transformer override">

    def visit_TheoryAtom(self, atom):
        if atom.term.name == "tel":
            return atom
        if atom.term.name == "show":
            self._map_show_atom(atom)

        def has_valid_condition(lit):
            return 'condition' in lit and lit['condition'] is not None and \
                lit['condition'] != []
        if self._in_head:
            iterable, literals, deontic_conditional = self._map_deontic_atom_with_sequence(atom)
            if deontic_conditional is not None:
                self._deontic_conditional = deontic_conditional
            if not iterable:
                return atom
            conditional_literals = [ast.ConditionalLiteral(atom.term.location, lit['term'], lit['condition'] or [])
                                    for lit in literals]
            self._head_theory_atoms_sequence.extend(conditional_literals)
        elif self._in_body:
            iterable, literals, deontic_conditional = self._map_deontic_atom_with_sequence(atom)
            if deontic_conditional is not None:
                raise Exception(f"Deontic conditional '{deontic_conditional}' not allowed in body")
            if not iterable:
                return atom
            rewritten_literals = self._rewrite_literals(literals)
            conditional_literals = [ast.ConditionalLiteral(atom.term.location, lit['term'], lit['condition'])
                                    if has_valid_condition(lit) else lit['term']
                                    for lit in rewritten_literals]
            self._body_theory_atoms_sequence.extend(conditional_literals)
        return atom

    # </editor-fold>

    # <editor-fold desc="Private methods">

    def _rewrite_literals(self, literals):
        new_literals = []
        for lit in literals:
            new_term = lit['term']
            new_condition = lit['condition'] if 'condition' in lit else None
            if new_term.atom.symbol.name in [DeonticAtoms.VIOLATED_OBLIGATION.value.prefixed(),
                                             DeonticAtoms.FULFILLED_OBLIGATION.value.prefixed(),
                                             DeonticAtoms.NON_VIOLATED_OBLIGATION.value.prefixed(),
                                             DeonticAtoms.NON_FULFILLED_OBLIGATION.value.prefixed(),
                                             DeonticAtoms.UNDETERMINED_OBLIGATION.value.prefixed()]:
                if new_term.sign == ast.Sign.Negation:
                    new_literals.append(lit)
                    self._derivation_rules_to_be_added.add(new_term.atom.symbol.name)
                    continue
                is_fulfillment = new_term.atom.symbol.name in [DeonticAtoms.FULFILLED_OBLIGATION.value.prefixed(),
                                                               DeonticAtoms.NON_FULFILLED_OBLIGATION.value.prefixed()]
                is_negated = new_term.atom.symbol.name in [DeonticAtoms.NON_VIOLATED_OBLIGATION.value.prefixed(),
                                                           DeonticAtoms.NON_FULFILLED_OBLIGATION.value.prefixed()]
                is_undetermined = new_term.atom.symbol.name == DeonticAtoms.UNDETERMINED_OBLIGATION.value.prefixed()
                new_term_2 = copy.deepcopy(new_term)
                new_term_3 = copy.deepcopy(new_term)
                new_term.atom.symbol.name = DeonticAtoms.OBLIGATORY.value.prefixed()
                new_term_2.atom.symbol.name = DeonticAtoms.HOLDS.value.prefixed()
                new_term_3.atom.symbol.name = DeonticAtoms.HOLDS.value.prefixed()
                new_literals.append({'term': new_term, 'condition': new_condition})
                if is_negated or is_undetermined:
                    new_term_2.sign = ast.Sign.Negation
                    new_term_3.sign = ast.Sign.Negation
                if not is_fulfillment and not is_undetermined:
                    self._negate_deontic_atom_content(new_term_2)
                new_literals.append({'term': new_term_2, 'condition': new_condition})
                if is_undetermined:
                    self._negate_deontic_atom_content(new_term_3)
                    new_literals.append({'term': new_term_3, 'condition': new_condition})
            elif new_term.atom.symbol.name in [DeonticAtoms.VIOLATED_PROHIBITION.value.prefixed(),
                                               DeonticAtoms.FULFILLED_PROHIBITION.value.prefixed(),
                                               DeonticAtoms.NON_VIOLATED_PROHIBITION.value.prefixed(),
                                               DeonticAtoms.NON_FULFILLED_PROHIBITION.value.prefixed(),
                                               DeonticAtoms.UNDETERMINED_PROHIBITION.value.prefixed()]:
                if new_term.sign == ast.Sign.Negation:
                    new_literals.append(lit)
                    self._derivation_rules_to_be_added.add(new_term.atom.symbol.name)
                    continue
                is_fulfillment = new_term.atom.symbol.name in [DeonticAtoms.FULFILLED_PROHIBITION.value.prefixed(),
                                                               DeonticAtoms.NON_FULFILLED_PROHIBITION.value.prefixed()]
                is_negated = new_term.atom.symbol.name in [DeonticAtoms.NON_VIOLATED_PROHIBITION.value.prefixed(),
                                                           DeonticAtoms.NON_FULFILLED_PROHIBITION.value.prefixed()]
                new_term_2 = copy.deepcopy(new_term)
                new_term.atom.symbol.name = DeonticAtoms.FORBIDDEN.value.prefixed()
                new_term_2.atom.symbol.name = DeonticAtoms.HOLDS.value.prefixed()
                new_literals.append({'term': new_term, 'condition': new_condition})
                if is_negated:
                    new_term_2.sign = ast.Sign.Negation
                if is_fulfillment:
                    self._negate_deontic_atom_content(new_term_2)
                new_literals.append({'term': new_term_2, 'condition': new_condition})
            elif new_term.atom.symbol.name in [DeonticAtoms.PERMITTED_IMPLICITLY.value.prefixed()]:
                new_term.atom.symbol.name = DeonticAtoms.FORBIDDEN.value.prefixed()
                new_term.sign = ast.Sign.Negation if new_term.sign == ast.Sign.NoSign else ast.Sign.DoubleNegation
                new_literals.append({'term': new_term, 'condition': new_condition})
            elif new_term.atom.symbol.name in [DeonticAtoms.OMISSIBLE_IMPLICITLY.value.prefixed()]:
                new_term.atom.symbol.name = DeonticAtoms.OBLIGATORY.value.prefixed()
                new_term.sign = ast.Sign.Negation if new_term.sign == ast.Sign.NoSign else ast.Sign.DoubleNegation
                new_literals.append({'term': new_term, 'condition': new_condition})
            else:
                new_literals.append(lit)
        return new_literals

    def _process_theory_atoms_in_head(self, multi_rule, new_head, rule):
        if (new_head.ast_type == ast.ASTType.Disjunction or new_head.ast_type == ast.ASTType.TheoryAtom) and \
               not self._is_telingo_atom(new_head):
            new_head_elements = self._filter_out_theory_atoms_and_literals_of_theory_atoms(
                new_head.elements) if new_head.ast_type == ast.ASTType.Disjunction else []
            if self._deontic_conditional is not None:
                new_rule = ast.Rule(rule.location,
                                    self._deontic_conditional.term,
                                    [self._deontic_conditional.condition])
                new_rule_2 = copy.deepcopy(new_rule)
                ob_atom_name = DeonticAtoms.OBLIGATORY.value.prefixed()
                nh_atom_name = DeonticAtoms.HOLDS.value.prefixed()
                conditional_term = self._deontic_conditional.condition.atom.symbol
                nh_conditional_term = copy.deepcopy(conditional_term)
                negative = "-"
                if str(nh_conditional_term).startswith("-"):
                    negative = ""
                    nh_conditional_term = nh_conditional_term.argument
                ob_atom = _symbolic_atom(rule.location, ob_atom_name, [conditional_term])
                nh_atom = _symbolic_atom(rule.location, negative + nh_atom_name, [nh_conditional_term])
                new_rule_2.body = [_positive_literal(rule.location, ob_atom),
                                   _negative_literal(rule.location, nh_atom)]
                multi_rule = [new_rule, new_rule_2]
            else:
                new_head_elements.extend(self._head_theory_atoms_sequence)
                new_head = ast.Disjunction(new_head.location, new_head_elements)
        return multi_rule, new_head

    def _negate_deontic_atom_content(self, deontic_atom):
        if deontic_atom.atom.symbol.ast_type == ast.ASTType.Function:
            if deontic_atom.atom.symbol.arguments[0].ast_type == ast.ASTType.UnaryOperation:
                deontic_atom.atom.symbol.arguments[0] = deontic_atom.atom.symbol.arguments[0].argument
            else:
                arg0 = deontic_atom.atom.symbol.arguments[0]
                deontic_atom.atom.symbol.arguments[0] = \
                    ast.UnaryOperation(deontic_atom.location, ast.UnaryOperator.Minus, arg0)
        else:
            deontic_atom.atom.symbol.name = "-" + deontic_atom.atom.symbol.name

    # </editor-fold>




from setuptools import setup, find_packages

version = {}
exec(open('deolingo/_version.py').read(), version)

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='deolingo',
    version=version['__version__'],
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'cffi>=1.16.0',
        'clingo>=5.7.1',
        'clingox>=1.2.1',
        'pycparser>=2.21',
        'prettytable>=2.2.0',
        'xclingo~=2.0b12',
        'langchain~=0.1.0',
        'openai~=1.30.1',
        'langchain-openai~=0.1.7',
        'gpt4all~=2.6.0',
        'transformers~=4.40.2',
        'google-generativeai~=0.5.3',
        'telingo~=2.1.2',
        'python-dotenv~=1.0.1'
    ],
    entry_points={
        'console_scripts': [
            'deolingo=deolingo.__main__:main'
        ]
    },
    author='Ovidio M. Moar',
    author_email='ovidio.manteiga@udc.es',
    description='A deontic logic solver for explainable deontic reasoning with Answer Set Programming (ASP).',
    license='MIT',
    keywords='ASP, NMR, Deontic Logic, SDL, Deontic Reasoning, Clingo, Xclingo, Telingo',
    url='https://github.com/ovidiomanteiga/deolingo',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Legal Industry',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
        'Topic :: Software Development :: Interpreters',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities'
    ]
)




from deolingo._ast_rewriting_transformer import DeonticASTRewritingTransformer

from deolingo._deolingo_theory import _DEOLINGO_RESTRICTED_THEORY
from deolingo._deontic_atom import DeonticAtoms
from deolingo._deontic_rules import DeonticRules
from deolingo._translator import DeolingoTranslator


class DeolingoRewritingTranslator(DeolingoTranslator):

    # <editor-fold desc="Initialization">

    def __init__(self, add_to_program_callback, transformer=None, translate=False,
                 add_theory=True, add_deontic_rules=True, weak=False):
        transformer = transformer if transformer is not None else DeonticASTRewritingTransformer(translate)
        super().__init__(add_to_program_callback, transformer, translate, add_theory, add_deontic_rules, weak=weak)

    # </editor-fold>

    # <editor-fold desc="Private methods"

    def _add_deolingo_theory(self):
        """Add the Deolingo theory to the program if it has not been added yet."""
        if self._add_theory and not self._deolingo_theory_added:
            self._add_string_to_program(_DEOLINGO_RESTRICTED_THEORY, add_to_translation=not self.translate)
            self._deolingo_theory_added = True

    def _add_rules_for_each_deontic_atom(self):
        super()._add_rules_for_each_deontic_atom()

    def _add_common_deontic_rules(self):
        if self._deontic_rules_added:
            return
        self._deontic_rules_added = True
        necessary_rules = self.deontic_transformer.derivation_rules_to_be_added
        if not self.weak:
            super()._add_string_to_program("\n".join(DeonticRules.deontic_weak_axiom()) + '\n')
        else:
            super()._add_string_to_program("\n".join(DeonticRules.deontic_weak_axiom_weak_constraint()) + '\n')
        super()._add_string_to_program("\n".join(DeonticRules.obligation_prohibition_equivalence()) + '\n')
        if len(necessary_rules) == 0:
            return
        super()._add_string_to_program("\n% Deontic rules\n")
        rules = {
            DeonticAtoms.VIOLATED_OBLIGATION.value.prefixed(): DeonticRules.derive_violated_obligation(),
            DeonticAtoms.FULFILLED_OBLIGATION.value.prefixed(): DeonticRules.derive_fulfilled_obligation(),
            DeonticAtoms.VIOLATED_PROHIBITION.value.prefixed(): DeonticRules.derive_violated_prohibition(),
            DeonticAtoms.FULFILLED_PROHIBITION.value.prefixed(): DeonticRules.derive_fulfilled_prohibition(),
            DeonticAtoms.NON_VIOLATED_OBLIGATION.value.prefixed(): DeonticRules.derive_non_violated_obligation(),
            DeonticAtoms.NON_FULFILLED_OBLIGATION.value.prefixed(): DeonticRules.derive_non_fulfilled_obligation(),
            DeonticAtoms.NON_VIOLATED_PROHIBITION.value.prefixed(): DeonticRules.derive_non_violated_prohibition(),
            DeonticAtoms.NON_FULFILLED_PROHIBITION.value.prefixed(): DeonticRules.derive_non_fulfilled_prohibition(),
            DeonticAtoms.UNDETERMINED_OBLIGATION.value.prefixed(): DeonticRules.derive_undetermined_obligation(),
            DeonticAtoms.UNDETERMINED_PROHIBITION.value.prefixed(): DeonticRules.derive_undetermined_prohibition(),
        }
        for deontic_atom_name in necessary_rules:
            super()._add_string_to_program(rules[deontic_atom_name] + "\n")

    # </editor-fold>




from deolingo._deontic_atom import *


class DeonticRules:

    @staticmethod
    def all_rules(weak=False):
        weakAxiom = DeonticRules.deontic_weak_axiom_weak_constraint() if weak else DeonticRules.deontic_weak_axiom()
        return [
            *weakAxiom,
            *DeonticRules.violation_fullfillment(),
            *DeonticRules.obligation_prohibition_equivalence(),
            *DeonticRules.implicit_permission_omission(),
            *DeonticRules.permissible_omissible_rules(),
            *DeonticRules.optional_definition(),
            *DeonticRules.permitted_omissible_defaults(),
            *DeonticRules.obligation_violation_fulfillment(),
            *DeonticRules.default_obligation(),
            *DeonticRules.prohibition_violation_fulfillment(),
            *DeonticRules.default_prohibition(),
            *DeonticRules.deontic_rules(),
        ]

    @staticmethod
    def all_rules_as_string(weak=False):
        return "\n".join(DeonticRules.all_rules(weak)) + '\n'

    @staticmethod
    def deontic_weak_axiom():
        return [
            f"\n% Deontic weak axiom D for DELX",
            f":- {obligatory('X')}, {forbidden('X')}, not {holds('X')}, not {holds('-X')}.",
        ]

    @staticmethod
    def deontic_weak_axiom_weak_constraint():
        return [
            "\n% Deontic weak axiom D for DELX (weak constraint)",
            f"deolingo_inconsistency(X) :- not deolingo_inconsistency(-X),"
                f" {obligatory('X')}, {forbidden('X')}, not {holds('X')}, not {holds('-X')}.",
            ":~ deolingo_inconsistency(X). [1@1, X]",
            "#show deolingo_inconsistency/1.",
            '%!trace {"INCONSISTENCY: % is obligatory and forbidden without factual information!", X} '
                'deolingo_inconsistency(X).\n',
            "%!show_trace deolingo_inconsistency(X).",
        ]

    @staticmethod
    def violation_fullfillment():
        return [
            f"\n% Violation",
            f"{violated('X')} :- {violated_obligation('X')}.",
            f"{violated('X')} :- {violated_prohibition('X')}.",
            f"\n% Fulfillment",
            f"{fulfilled('X')} :- {fulfilled_obligation('X')}.",
            f"{fulfilled('X')} :- {fulfilled_prohibition('X')}.",
        ]

    @staticmethod
    def obligation_prohibition_equivalence():
        return [
            f"\n% Obligation/prohibition equivalence",
            f"{obligatory('X')} :- {forbidden('-X')}.",
            f"{forbidden('X')}  :- {obligatory('-X')}.",
            f"-{obligatory('X')} :- -{forbidden('-X')}.",
            f"-{forbidden('X')}  :- -{obligatory('-X')}.",
        ]

    @staticmethod
    def implicit_permission_omission():
        return [
            f"\n% Implicit permission",
            f"{permitted_implicitly('X')} :- not {forbidden('X')}, {deontic('X')}.",
            f"not {forbidden('X')} :- {permitted_implicitly('X')}.",
            f"\n% Implicit omission",
            f"{omissible_implicitly('X')} :- not {obligatory('X')}, {deontic('X')}.",
            f"not {obligatory('X')} :- {omissible_implicitly('X')}.",
        ]

    @staticmethod
    def permissible_omissible_rules():
        return [
            f"\n% Permissible and omissible",
            f"{omissible('X')} :- -{obligatory('X')}.",
            f"-{obligatory('X')} :- {omissible('X')}.",
            f"{permitted('X')} :- -{forbidden('X')}.",
            f"-{forbidden('X')} :- {permitted('X')}.",
            f"-{omissible('X')} :- {obligatory('X')}.",
            f"{obligatory('X')} :- -{omissible('X')}.",
            f"-{permitted('X')} :- {forbidden('X')}.",
            f"{forbidden('X')} :- -{permitted('X')}.",
        ]

    @staticmethod
    def optional_definition():
        return [
            f"\n% Optional",
            f"{optional('X')} :- {omissible('X')}, {permitted('X')}.",
            f"{omissible('X')} :- {optional('X')}.",
            f"{permitted('X')} :- {optional('X')}.",
        ]

    @staticmethod
    def permitted_omissible_defaults():
        return [
            f"\n% Permitted by default",
            f"{permitted_by_default('X')} :- not not {forbidden('X')}; {deontic('X')}.",
            f"{permitted_by_default('X')} :- {permitted('X')}.",
            f"{permitted('X')} :- not {forbidden('X')}, {permitted_by_default('X')}.",
            f"\n% Omissible by default",
            f"{omissible_by_default('X')} :- not not {obligatory('X')}; {deontic('X')}.",
            f"{omissible_by_default('X')} :- {omissible('X')}.",
            f"{omissible('X')} :- not {obligatory('X')}, {omissible_by_default('X')}.",
        ]

    @staticmethod
    def obligation_violation_fulfillment():
        return [
            f"\n% Obligation violation",
            DeonticRules.derive_violated_obligation(),
            f"{obligatory('X')} :- {violated_obligation('X')}.",
            f"{holds('-X')} :- {violated_obligation('X')}.",
            f"\n% Fulfilled obligation",
            DeonticRules.derive_fulfilled_obligation(),
            f"{obligatory('X')} :- {fulfilled_obligation('X')}.",
            f"{holds('X')} :- {fulfilled_obligation('X')}.",
            f"\n% Non-fulfilled obligation",
            DeonticRules.derive_non_fulfilled_obligation(),
            f"{obligatory('X')} :- {non_fulfilled_obligation('X')}.",
            f"not {holds('X')} :- {non_fulfilled_obligation('X')}.",
            f"\n% Non-violated obligation",
            DeonticRules.derive_non_violated_obligation(),
            f"{obligatory('X')} :- {non_violated_obligation('X')}.",
            f"not {holds('-X')} :- {non_violated_obligation('X')}.",
            f"\n% Undetermined obligation",
            DeonticRules.derive_undetermined_obligation(),
            f"{obligatory('X')} :- {undetermined_obligation('X')}.",
            f"not {holds('X')} :- {undetermined_obligation('X')}.",
            f"not {holds('-X')} :- {undetermined_obligation('X')}.",
        ]

    @staticmethod
    def derive_violated_obligation():
        return f"{violated_obligation('X')} :- {obligatory('X')}, {holds('-X')}."

    @staticmethod
    def derive_fulfilled_obligation():
        return f"{fulfilled_obligation('X')} :- {obligatory('X')}, {holds('X')}."

    @staticmethod
    def derive_non_fulfilled_obligation():
        return f"{non_fulfilled_obligation('X')} :- {obligatory('X')}, not {holds('X')}."

    @staticmethod
    def derive_non_violated_obligation():
        return f"{non_violated_obligation('X')} :- {obligatory('X')}, not {holds('-X')}."

    @staticmethod
    def derive_undetermined_obligation():
        return f"{undetermined_obligation('X')} :- {obligatory('X')}, not {holds('X')}, not {holds('-X')}."

    @staticmethod
    def default_obligation():
        return [
            f"\n% Default obligation",
            f"{default_obligation('X')} :- not not {permitted('-X')}; {deontic('X')}.",
            f"{default_obligation('X')} :- {obligatory('X')}.",
            f"{obligatory('X')} :- not {permitted('-X')}, {default_obligation('X')}.",
        ]

    @staticmethod
    def prohibition_violation_fulfillment():
        return [
            f"\n% Violated prohibition",
            DeonticRules.derive_violated_prohibition(),
            f"{forbidden('X')} :- {violated_prohibition('X')}.",
            f"{holds('X')} :- {violated_prohibition('X')}.",
            f"\n% Fulfilled prohibition",
            DeonticRules.derive_fulfilled_prohibition(),
            f"{forbidden('X')} :- {fulfilled_prohibition('X')}.",
            f"{holds('-X')} :- {fulfilled_prohibition('X')}.",
            f"\n% Non-fulfilled prohibition",
            DeonticRules.derive_non_fulfilled_prohibition(),
            f"{forbidden('X')} :- {non_fulfilled_prohibition('X')}.",
            f"not {holds('-X')} :- {non_fulfilled_prohibition('X')}.",
            f"\n% Non-violated prohibition",
            DeonticRules.derive_non_violated_prohibition(),
            f"{forbidden('X')} :- {non_violated_prohibition('X')}.",
            f"not {holds('X')} :- {non_violated_prohibition('X')}.",
            f"\n% Undetermined prohibition",
            DeonticRules.derive_undetermined_prohibition(),
            f"{forbidden('X')} :- {undetermined_prohibition('X')}.",
            f"not {holds('X')} :- {undetermined_prohibition('X')}.",
            f"not {holds('-X')} :- {undetermined_prohibition('X')}.",
        ]

    # same as above but with prohibition

    @staticmethod
    def derive_violated_prohibition():
        return f"{violated_prohibition('X')} :- {forbidden('X')}, {holds('X')}."

    @staticmethod
    def derive_fulfilled_prohibition():
        return f"{fulfilled_prohibition('X')} :- {forbidden('X')}, {holds('-X')}."

    @staticmethod
    def derive_non_fulfilled_prohibition():
        return f"{non_fulfilled_prohibition('X')} :- {forbidden('X')}, not {holds('-X')}."

    @staticmethod
    def derive_non_violated_prohibition():
        return f"{non_violated_prohibition('X')} :- {forbidden('X')}, not {holds('X')}."

    @staticmethod
    def derive_undetermined_prohibition():
        return f"{undetermined_prohibition('X')} :- {forbidden('X')}, not {holds('X')}, not {holds('-X')}."

    @staticmethod
    def default_prohibition():
        return [
            f"\n% Default prohibition",
            f"{default_prohibition('X')} :- not not {permitted('X')}; {deontic('X')}.",
            f"{default_prohibition('X')} :- {forbidden('X')}.",
            f"{forbidden('X')} :- not {permitted('X')}, {default_prohibition('X')}.",
        ]

    @staticmethod
    def deontic_rules():
        return [
            f"\n% Deontic",
            f"{deontic('X')} :- {obligatory('X')}.",
            f"{deontic('X')} :- {deontic('-X')}.",
        ]
